"""
Tests for salt.utils.data
"""

import builtins
import logging

import salt.utils.data
import salt.utils.stringutils
from salt.utils.odict import OrderedDict
from tests.support.mock import patch
from tests.support.unit import LOREM_IPSUM, TestCase

log = logging.getLogger(__name__)
_b = lambda x: x.encode("utf-8")
_s = lambda x: salt.utils.stringutils.to_str(x, normalize=True)
# Some randomized data that will not decode
BYTES = b"1\x814\x10"

# This is an example of a unicode string with й constructed using two separate
# code points. Do not modify it.
EGGS = "\u044f\u0438\u0306\u0446\u0430"


class DataTestCase(TestCase):
    test_data = [
        "unicode_str",
        _b("питон"),
        123,
        456.789,
        True,
        False,
        None,
        EGGS,
        BYTES,
        [123, 456.789, _b("спам"), True, False, None, EGGS, BYTES],
        (987, 654.321, _b("яйца"), EGGS, None, (True, EGGS, BYTES)),
        {
            _b("str_key"): _b("str_val"),
            None: True,
            123: 456.789,
            EGGS: BYTES,
            _b("subdict"): {
                "unicode_key": EGGS,
                _b("tuple"): (123, "hello", _b("world"), True, EGGS, BYTES),
                _b("list"): [456, _b("спам"), False, EGGS, BYTES],
            },
        },
        OrderedDict([(_b("foo"), "bar"), (123, 456), (EGGS, BYTES)]),
    ]

    def test_sorted_ignorecase(self):
        test_list = ["foo", "Foo", "bar", "Bar"]
        expected_list = ["bar", "Bar", "foo", "Foo"]
        self.assertEqual(salt.utils.data.sorted_ignorecase(test_list), expected_list)

    def test_mysql_to_dict(self):
        test_mysql_output = [
            "+----+------+-----------+------+---------+------+-------+------------------+",
            "| Id | User | Host      | db   | Command | Time | State | Info         "
            "    |",
            "+----+------+-----------+------+---------+------+-------+------------------+",
            "|  7 | root | localhost | NULL | Query   |    0 | init  | show"
            " processlist |",
            "+----+------+-----------+------+---------+------+-------+------------------+",
        ]

        ret = salt.utils.data.mysql_to_dict(test_mysql_output, "Info")
        expected_dict = {
            "show processlist": {
                "Info": "show processlist",
                "db": "NULL",
                "State": "init",
                "Host": "localhost",
                "Command": "Query",
                "User": "root",
                "Time": 0,
                "Id": 7,
            }
        }

        self.assertDictEqual(ret, expected_dict)

    def test_subdict_match(self):
        test_two_level_dict = {"foo": {"bar": "baz"}}
        test_two_level_comb_dict = {"foo": {"bar": "baz:woz"}}
        test_two_level_dict_and_list = {
            "abc": ["def", "ghi", {"lorem": {"ipsum": [{"dolor": "sit"}]}}],
        }
        test_three_level_dict = {"a": {"b": {"c": "v"}}}

        self.assertTrue(
            salt.utils.data.subdict_match(test_two_level_dict, "foo:bar:baz")
        )
        # In test_two_level_comb_dict, 'foo:bar' corresponds to 'baz:woz', not
        # 'baz'. This match should return False.
        self.assertFalse(
            salt.utils.data.subdict_match(test_two_level_comb_dict, "foo:bar:baz")
        )
        # This tests matching with the delimiter in the value part (in other
        # words, that the path 'foo:bar' corresponds to the string 'baz:woz').
        self.assertTrue(
            salt.utils.data.subdict_match(test_two_level_comb_dict, "foo:bar:baz:woz")
        )
        # This would match if test_two_level_comb_dict['foo']['bar'] was equal
        # to 'baz:woz:wiz', or if there was more deep nesting. But it does not,
        # so this should return False.
        self.assertFalse(
            salt.utils.data.subdict_match(
                test_two_level_comb_dict, "foo:bar:baz:woz:wiz"
            )
        )
        # This tests for cases when a key path corresponds to a list. The
        # value part 'ghi' should be successfully matched as it is a member of
        # the list corresponding to key path 'abc'. It is somewhat a
        # duplication of a test within test_traverse_dict_and_list, but
        # salt.utils.data.subdict_match() does more than just invoke
        # salt.utils.traverse_list_and_dict() so this particular assertion is a
        # sanity check.
        self.assertTrue(
            salt.utils.data.subdict_match(test_two_level_dict_and_list, "abc:ghi")
        )
        # This tests the use case of a dict embedded in a list, embedded in a
        # list, embedded in a dict. This is a rather absurd case, but it
        # confirms that match recursion works properly.
        self.assertTrue(
            salt.utils.data.subdict_match(
                test_two_level_dict_and_list, "abc:lorem:ipsum:dolor:sit"
            )
        )
        # Test four level dict match for reference
        self.assertTrue(salt.utils.data.subdict_match(test_three_level_dict, "a:b:c:v"))
        # Test regression in 2015.8 where 'a:c:v' would match 'a:b:c:v'
        self.assertFalse(salt.utils.data.subdict_match(test_three_level_dict, "a:c:v"))
        # Test wildcard match
        self.assertTrue(salt.utils.data.subdict_match(test_three_level_dict, "a:*:c:v"))

    def test_subdict_match_with_wildcards(self):
        """
        Tests subdict matching when wildcards are used in the expression
        """
        data = {"a": {"b": {"ç": "d", "é": ["eff", "gee", "8ch"], "ĩ": {"j": "k"}}}}
        assert salt.utils.data.subdict_match(data, "*:*:*:*")
        assert salt.utils.data.subdict_match(data, "a:*:*:*")
        assert salt.utils.data.subdict_match(data, "a:b:*:*")
        assert salt.utils.data.subdict_match(data, "a:b:ç:*")
        assert salt.utils.data.subdict_match(data, "a:b:*:d")
        assert salt.utils.data.subdict_match(data, "a:*:ç:d")
        assert salt.utils.data.subdict_match(data, "*:b:ç:d")
        assert salt.utils.data.subdict_match(data, "*:*:ç:d")
        assert salt.utils.data.subdict_match(data, "*:*:*:d")
        assert salt.utils.data.subdict_match(data, "a:*:*:d")
        assert salt.utils.data.subdict_match(data, "a:b:*:ef*")
        assert salt.utils.data.subdict_match(data, "a:b:*:g*")
        assert salt.utils.data.subdict_match(data, "a:b:*:j:*")
        assert salt.utils.data.subdict_match(data, "a:b:*:j:k")
        assert salt.utils.data.subdict_match(data, "a:b:*:*:k")
        assert salt.utils.data.subdict_match(data, "a:b:*:*:*")

    def test_traverse_dict(self):
        test_two_level_dict = {"foo": {"bar": "baz"}}

        self.assertDictEqual(
            {"not_found": "nope"},
            salt.utils.data.traverse_dict(
                test_two_level_dict, "foo:bar:baz", {"not_found": "nope"}
            ),
        )
        self.assertEqual(
            "baz",
            salt.utils.data.traverse_dict(
                test_two_level_dict, "foo:bar", {"not_found": "not_found"}
            ),
        )

    def test_traverse_dict_and_list(self):
        test_two_level_dict = {"foo": {"bar": "baz"}}
        test_two_level_dict_and_list = {
            "foo": ["bar", "baz", {"lorem": {"ipsum": [{"dolor": "sit"}]}}]
        }

        # Check traversing too far: salt.utils.data.traverse_dict_and_list() returns
        # the value corresponding to a given key path, and baz is a value
        # corresponding to the key path foo:bar.
        self.assertDictEqual(
            {"not_found": "nope"},
            salt.utils.data.traverse_dict_and_list(
                test_two_level_dict, "foo:bar:baz", {"not_found": "nope"}
            ),
        )
        # Now check to ensure that foo:bar corresponds to baz
        self.assertEqual(
            "baz",
            salt.utils.data.traverse_dict_and_list(
                test_two_level_dict, "foo:bar", {"not_found": "not_found"}
            ),
        )
        # Check traversing too far
        self.assertDictEqual(
            {"not_found": "nope"},
            salt.utils.data.traverse_dict_and_list(
                test_two_level_dict_and_list, "foo:bar", {"not_found": "nope"}
            ),
        )
        # Check index 1 (2nd element) of list corresponding to path 'foo'
        self.assertEqual(
            "baz",
            salt.utils.data.traverse_dict_and_list(
                test_two_level_dict_and_list, "foo:1", {"not_found": "not_found"}
            ),
        )
        # Traverse a couple times into dicts embedded in lists
        self.assertEqual(
            "sit",
            salt.utils.data.traverse_dict_and_list(
                test_two_level_dict_and_list,
                "foo:lorem:ipsum:dolor",
                {"not_found": "not_found"},
            ),
        )

        # Traverse and match integer key in a nested dict
        # https://github.com/saltstack/salt/issues/56444
        self.assertEqual(
            "it worked",
            salt.utils.data.traverse_dict_and_list(
                {"foo": {1234: "it worked"}},
                "foo:1234",
                "it didn't work",
            ),
        )
        # Make sure that we properly return the default value when the initial
        # attempt fails and YAML-loading the target key doesn't change its
        # value.
        self.assertEqual(
            "default",
            salt.utils.data.traverse_dict_and_list(
                {"foo": {"baz": "didn't work"}},
                "foo:bar",
                "default",
            ),
        )

    def test_issue_39709(self):
        test_two_level_dict_and_list = {
            "foo": ["bar", "baz", {"lorem": {"ipsum": [{"dolor": "sit"}]}}]
        }

        self.assertEqual(
            "sit",
            salt.utils.data.traverse_dict_and_list(
                test_two_level_dict_and_list,
                ["foo", "lorem", "ipsum", "dolor"],
                {"not_found": "not_found"},
            ),
        )

    def test_compare_dicts(self):
        ret = salt.utils.data.compare_dicts(old={"foo": "bar"}, new={"foo": "bar"})
        self.assertEqual(ret, {})

        ret = salt.utils.data.compare_dicts(old={"foo": "bar"}, new={"foo": "woz"})
        expected_ret = {"foo": {"new": "woz", "old": "bar"}}
        self.assertDictEqual(ret, expected_ret)

    def test_compare_lists_no_change(self):
        ret = salt.utils.data.compare_lists(
            old=[1, 2, 3, "a", "b", "c"], new=[1, 2, 3, "a", "b", "c"]
        )
        expected = {}
        self.assertDictEqual(ret, expected)

    def test_compare_lists_changes(self):
        ret = salt.utils.data.compare_lists(
            old=[1, 2, 3, "a", "b", "c"], new=[1, 2, 4, "x", "y", "z"]
        )
        expected = {"new": [4, "x", "y", "z"], "old": [3, "a", "b", "c"]}
        self.assertDictEqual(ret, expected)

    def test_compare_lists_changes_new(self):
        ret = salt.utils.data.compare_lists(old=[1, 2, 3], new=[1, 2, 3, "x", "y", "z"])
        expected = {"new": ["x", "y", "z"]}
        self.assertDictEqual(ret, expected)

    def test_compare_lists_changes_old(self):
        ret = salt.utils.data.compare_lists(old=[1, 2, 3, "a", "b", "c"], new=[1, 2, 3])
        expected = {"old": ["a", "b", "c"]}
        self.assertDictEqual(ret, expected)

    def test_decode(self):
        """
        Companion to test_decode_to_str, they should both be kept up-to-date
        with one another.

        NOTE: This uses the lambda "_b" defined above in the global scope,
        which encodes a string to a bytestring, assuming utf-8.
        """
        expected = [
            "unicode_str",
            "питон",
            123,
            456.789,
            True,
            False,
            None,
            "яйца",
            BYTES,
            [123, 456.789, "спам", True, False, None, "яйца", BYTES],
            (987, 654.321, "яйца", "яйца", None, (True, "яйца", BYTES)),
            {
                "str_key": "str_val",
                None: True,
                123: 456.789,
                "яйца": BYTES,
                "subdict": {
                    "unicode_key": "яйца",
                    "tuple": (123, "hello", "world", True, "яйца", BYTES),
                    "list": [456, "спам", False, "яйца", BYTES],
                },
            },
            OrderedDict([("foo", "bar"), (123, 456), ("яйца", BYTES)]),
        ]

        ret = salt.utils.data.decode(
            self.test_data,
            keep=True,
            normalize=True,
            preserve_dict_class=True,
            preserve_tuples=True,
        )
        self.assertEqual(ret, expected)

        # The binary data in the data structure should fail to decode, even
        # using the fallback, and raise an exception.
        self.assertRaises(
            UnicodeDecodeError,
            salt.utils.data.decode,
            self.test_data,
            keep=False,
            normalize=True,
            preserve_dict_class=True,
            preserve_tuples=True,
        )

        # Now munge the expected data so that we get what we would expect if we
        # disable preservation of dict class and tuples
        expected[10] = [987, 654.321, "яйца", "яйца", None, [True, "яйца", BYTES]]
        expected[11]["subdict"]["tuple"] = [123, "hello", "world", True, "яйца", BYTES]
        expected[12] = {"foo": "bar", 123: 456, "яйца": BYTES}

        ret = salt.utils.data.decode(
            self.test_data,
            keep=True,
            normalize=True,
            preserve_dict_class=False,
            preserve_tuples=False,
        )
        self.assertEqual(ret, expected)

        # Now test single non-string, non-data-structure items, these should
        # return the same value when passed to this function
        for item in (123, 4.56, True, False, None):
            log.debug("Testing decode of %s", item)
            self.assertEqual(salt.utils.data.decode(item), item)

        # Test single strings (not in a data structure)
        self.assertEqual(salt.utils.data.decode("foo"), "foo")
        self.assertEqual(salt.utils.data.decode(_b("bar")), "bar")
        self.assertEqual(salt.utils.data.decode(EGGS, normalize=True), "яйца")
        self.assertEqual(salt.utils.data.decode(EGGS, normalize=False), EGGS)

        # Test binary blob
        self.assertEqual(salt.utils.data.decode(BYTES, keep=True), BYTES)
        self.assertRaises(UnicodeDecodeError, salt.utils.data.decode, BYTES, keep=False)

    def test_circular_refs_dicts(self):
        test_dict = {"key": "value", "type": "test1"}
        test_dict["self"] = test_dict
        ret = salt.utils.data._remove_circular_refs(ob=test_dict)
        self.assertDictEqual(ret, {"key": "value", "type": "test1", "self": None})

    def test_circular_refs_lists(self):
        test_list = {
            "foo": [],
        }
        test_list["foo"].append((test_list,))
        ret = salt.utils.data._remove_circular_refs(ob=test_list)
        self.assertDictEqual(ret, {"foo": [(None,)]})

    def test_circular_refs_tuple(self):
        test_dup = {"foo": "string 1", "bar": "string 1", "ham": 1, "spam": 1}
        ret = salt.utils.data._remove_circular_refs(ob=test_dup)
        self.assertDictEqual(
            ret, {"foo": "string 1", "bar": "string 1", "ham": 1, "spam": 1}
        )

    def test_decode_to_str(self):
        """
        Companion to test_decode, they should both be kept up-to-date with one
        another.

        NOTE: This uses the lambda "_s" defined above in the global scope,
        which converts the string/bytestring to a str type.
        """
        expected = [
            _s("unicode_str"),
            _s("питон"),
            123,
            456.789,
            True,
            False,
            None,
            _s("яйца"),
            BYTES,
            [123, 456.789, _s("спам"), True, False, None, _s("яйца"), BYTES],
            (987, 654.321, _s("яйца"), _s("яйца"), None, (True, _s("яйца"), BYTES)),
            {
                _s("str_key"): _s("str_val"),
                None: True,
                123: 456.789,
                _s("яйца"): BYTES,
                _s("subdict"): {
                    _s("unicode_key"): _s("яйца"),
                    _s("tuple"): (
                        123,
                        _s("hello"),
                        _s("world"),
                        True,
                        _s("яйца"),
                        BYTES,
                    ),
                    _s("list"): [456, _s("спам"), False, _s("яйца"), BYTES],
                },
            },
            OrderedDict([(_s("foo"), _s("bar")), (123, 456), (_s("яйца"), BYTES)]),
        ]

        ret = salt.utils.data.decode(
            self.test_data,
            keep=True,
            normalize=True,
            preserve_dict_class=True,
            preserve_tuples=True,
            to_str=True,
        )
        self.assertEqual(ret, expected)

        # The binary data in the data structure should fail to decode, even
        # using the fallback, and raise an exception.
        self.assertRaises(
            UnicodeDecodeError,
            salt.utils.data.decode,
            self.test_data,
            keep=False,
            normalize=True,
            preserve_dict_class=True,
            preserve_tuples=True,
            to_str=True,
        )

        # Now munge the expected data so that we get what we would expect if we
        # disable preservation of dict class and tuples
        expected[10] = [
            987,
            654.321,
            _s("яйца"),
            _s("яйца"),
            None,
            [True, _s("яйца"), BYTES],
        ]
        expected[11][_s("subdict")][_s("tuple")] = [
            123,
            _s("hello"),
            _s("world"),
            True,
            _s("яйца"),
            BYTES,
        ]
        expected[12] = {_s("foo"): _s("bar"), 123: 456, _s("яйца"): BYTES}

        ret = salt.utils.data.decode(
            self.test_data,
            keep=True,
            normalize=True,
            preserve_dict_class=False,
            preserve_tuples=False,
            to_str=True,
        )
        self.assertEqual(ret, expected)

        # Now test single non-string, non-data-structure items, these should
        # return the same value when passed to this function
        for item in (123, 4.56, True, False, None):
            log.debug("Testing decode of %s", item)
            self.assertEqual(salt.utils.data.decode(item, to_str=True), item)

        # Test single strings (not in a data structure)
        self.assertEqual(salt.utils.data.decode("foo", to_str=True), _s("foo"))
        self.assertEqual(salt.utils.data.decode(_b("bar"), to_str=True), _s("bar"))

        # Test binary blob
        self.assertEqual(salt.utils.data.decode(BYTES, keep=True, to_str=True), BYTES)
        self.assertRaises(
            UnicodeDecodeError,
            salt.utils.data.decode,
            BYTES,
            keep=False,
            to_str=True,
        )

    def test_decode_fallback(self):
        """
        Test fallback to utf-8
        """
        with patch.object(builtins, "__salt_system_encoding__", "ascii"):
            self.assertEqual(salt.utils.data.decode(_b("яйца")), "яйца")

    def test_encode(self):
        """
        NOTE: This uses the lambda "_b" defined above in the global scope,
        which encodes a string to a bytestring, assuming utf-8.
        """
        expected = [
            _b("unicode_str"),
            _b("питон"),
            123,
            456.789,
            True,
            False,
            None,
            _b(EGGS),
            BYTES,
            [123, 456.789, _b("спам"), True, False, None, _b(EGGS), BYTES],
            (987, 654.321, _b("яйца"), _b(EGGS), None, (True, _b(EGGS), BYTES)),
            {
                _b("str_key"): _b("str_val"),
                None: True,
                123: 456.789,
                _b(EGGS): BYTES,
                _b("subdict"): {
                    _b("unicode_key"): _b(EGGS),
                    _b("tuple"): (123, _b("hello"), _b("world"), True, _b(EGGS), BYTES),
                    _b("list"): [456, _b("спам"), False, _b(EGGS), BYTES],
                },
            },
            OrderedDict([(_b("foo"), _b("bar")), (123, 456), (_b(EGGS), BYTES)]),
        ]

        # Both keep=True and keep=False should work because the BYTES data is
        # already bytes.
        ret = salt.utils.data.encode(
            self.test_data, keep=True, preserve_dict_class=True, preserve_tuples=True
        )
        self.assertEqual(ret, expected)
        ret = salt.utils.data.encode(
            self.test_data, keep=False, preserve_dict_class=True, preserve_tuples=True
        )
        self.assertEqual(ret, expected)

        # Now munge the expected data so that we get what we would expect if we
        # disable preservation of dict class and tuples
        expected[10] = [
            987,
            654.321,
            _b("яйца"),
            _b(EGGS),
            None,
            [True, _b(EGGS), BYTES],
        ]
        expected[11][_b("subdict")][_b("tuple")] = [
            123,
            _b("hello"),
            _b("world"),
            True,
            _b(EGGS),
            BYTES,
        ]
        expected[12] = {_b("foo"): _b("bar"), 123: 456, _b(EGGS): BYTES}

        ret = salt.utils.data.encode(
            self.test_data, keep=True, preserve_dict_class=False, preserve_tuples=False
        )
        self.assertEqual(ret, expected)
        ret = salt.utils.data.encode(
            self.test_data, keep=False, preserve_dict_class=False, preserve_tuples=False
        )
        self.assertEqual(ret, expected)

        # Now test single non-string, non-data-structure items, these should
        # return the same value when passed to this function
        for item in (123, 4.56, True, False, None):
            log.debug("Testing encode of %s", item)
            self.assertEqual(salt.utils.data.encode(item), item)

        # Test single strings (not in a data structure)
        self.assertEqual(salt.utils.data.encode("foo"), _b("foo"))
        self.assertEqual(salt.utils.data.encode(_b("bar")), _b("bar"))

        # Test binary blob, nothing should happen even when keep=False since
        # the data is already bytes
        self.assertEqual(salt.utils.data.encode(BYTES, keep=True), BYTES)
        self.assertEqual(salt.utils.data.encode(BYTES, keep=False), BYTES)

    def test_encode_keep(self):
        """
        Whereas we tested the keep argument in test_decode, it is much easier
        to do a more comprehensive test of keep in its own function where we
        can force the encoding.
        """
        unicode_str = "питон"
        encoding = "ascii"

        # Test single string
        self.assertEqual(
            salt.utils.data.encode(unicode_str, encoding, keep=True), unicode_str
        )
        self.assertRaises(
            UnicodeEncodeError,
            salt.utils.data.encode,
            unicode_str,
            encoding,
            keep=False,
        )

        data = [
            unicode_str,
            [b"foo", [unicode_str], {b"key": unicode_str}, (unicode_str,)],
            {
                b"list": [b"foo", unicode_str],
                b"dict": {b"key": unicode_str},
                b"tuple": (b"foo", unicode_str),
            },
            ([b"foo", unicode_str], {b"key": unicode_str}, (unicode_str,)),
        ]

        # Since everything was a bytestring aside from the bogus data, the
        # return data should be identical. We don't need to test recursive
        # decoding, that has already been tested in test_encode.
        self.assertEqual(
            salt.utils.data.encode(data, encoding, keep=True, preserve_tuples=True),
            data,
        )
        self.assertRaises(
            UnicodeEncodeError,
            salt.utils.data.encode,
            data,
            encoding,
            keep=False,
            preserve_tuples=True,
        )

        for index, _ in enumerate(data):
            self.assertEqual(
                salt.utils.data.encode(
                    data[index], encoding, keep=True, preserve_tuples=True
                ),
                data[index],
            )
            self.assertRaises(
                UnicodeEncodeError,
                salt.utils.data.encode,
                data[index],
                encoding,
                keep=False,
                preserve_tuples=True,
            )

    def test_encode_fallback(self):
        """
        Test fallback to utf-8
        """
        with patch.object(builtins, "__salt_system_encoding__", "ascii"):
            self.assertEqual(salt.utils.data.encode("яйца"), _b("яйца"))
        with patch.object(builtins, "__salt_system_encoding__", "CP1252"):
            self.assertEqual(salt.utils.data.encode("Ψ"), _b("Ψ"))

    def test_repack_dict(self):
        list_of_one_element_dicts = [
            {"dict_key_1": "dict_val_1"},
            {"dict_key_2": "dict_val_2"},
            {"dict_key_3": "dict_val_3"},
        ]
        expected_ret = {
            "dict_key_1": "dict_val_1",
            "dict_key_2": "dict_val_2",
            "dict_key_3": "dict_val_3",
        }
        ret = salt.utils.data.repack_dictlist(list_of_one_element_dicts)
        self.assertDictEqual(ret, expected_ret)

        # Try with yaml
        yaml_key_val_pair = "- key1: val1"
        ret = salt.utils.data.repack_dictlist(yaml_key_val_pair)
        self.assertDictEqual(ret, {"key1": "val1"})

        # Make sure we handle non-yaml junk data
        ret = salt.utils.data.repack_dictlist(LOREM_IPSUM)
        self.assertDictEqual(ret, {})

    def test_stringify(self):
        self.assertRaises(TypeError, salt.utils.data.stringify, 9)
        self.assertEqual(
            salt.utils.data.stringify(["one", "two", "three", 4, 5]),
            ["one", "two", "three", "4", "5"],
        )

    def test_json_query(self):
        # Raises exception if jmespath module is not found
        with patch("salt.utils.data.jmespath", None):
            self.assertRaisesRegex(
                RuntimeError, "requires jmespath", salt.utils.data.json_query, {}, "@"
            )

        # Test search
        user_groups = {
            "user1": {"groups": ["group1", "group2", "group3"]},
            "user2": {"groups": ["group1", "group2"]},
            "user3": {"groups": ["group3"]},
        }
        expression = "*.groups[0]"
        primary_groups = ["group1", "group1", "group3"]
        self.assertEqual(
            sorted(salt.utils.data.json_query(user_groups, expression)), primary_groups
        )


class FilterFalseyTestCase(TestCase):
    """
    Test suite for salt.utils.data.filter_falsey
    """

    def test_nop(self):
        """
        Test cases where nothing will be done.
        """
        # Test with dictionary without recursion
        old_dict = {
            "foo": "bar",
            "bar": {"baz": {"qux": "quux"}},
            "baz": ["qux", {"foo": "bar"}],
        }
        new_dict = salt.utils.data.filter_falsey(old_dict)
        self.assertEqual(old_dict, new_dict)
        # Check returned type equality
        self.assertIs(type(old_dict), type(new_dict))
        # Test dictionary with recursion
        new_dict = salt.utils.data.filter_falsey(old_dict, recurse_depth=3)
        self.assertEqual(old_dict, new_dict)
        # Test with list
        old_list = ["foo", "bar"]
        new_list = salt.utils.data.filter_falsey(old_list)
        self.assertEqual(old_list, new_list)
        # Check returned type equality
        self.assertIs(type(old_list), type(new_list))
        # Test with set
        old_set = {"foo", "bar"}
        new_set = salt.utils.data.filter_falsey(old_set)
        self.assertEqual(old_set, new_set)
        # Check returned type equality
        self.assertIs(type(old_set), type(new_set))
        # Test with OrderedDict
        old_dict = OrderedDict(
            [
                ("foo", "bar"),
                ("bar", OrderedDict([("qux", "quux")])),
                ("baz", ["qux", OrderedDict([("foo", "bar")])]),
            ]
        )
        new_dict = salt.utils.data.filter_falsey(old_dict)
        self.assertEqual(old_dict, new_dict)
        self.assertIs(type(old_dict), type(new_dict))
        # Test excluding int
        old_list = [0]
        new_list = salt.utils.data.filter_falsey(old_list, ignore_types=[type(0)])
        self.assertEqual(old_list, new_list)
        # Test excluding str (or unicode) (or both)
        old_list = [""]
        new_list = salt.utils.data.filter_falsey(old_list, ignore_types=[type("")])
        self.assertEqual(old_list, new_list)
        # Test excluding list
        old_list = [[]]
        new_list = salt.utils.data.filter_falsey(old_list, ignore_types=[type([])])
        self.assertEqual(old_list, new_list)
        # Test excluding dict
        old_list = [{}]
        new_list = salt.utils.data.filter_falsey(old_list, ignore_types=[type({})])
        self.assertEqual(old_list, new_list)

    def test_filter_dict_no_recurse(self):
        """
        Test filtering a dictionary without recursing.
        This will only filter out key-values where the values are falsey.
        """
        old_dict = {
            "foo": None,
            "bar": {"baz": {"qux": None, "quux": "", "foo": []}},
            "baz": ["qux"],
            "qux": {},
            "quux": [],
        }
        new_dict = salt.utils.data.filter_falsey(old_dict)
        expect_dict = {
            "bar": {"baz": {"qux": None, "quux": "", "foo": []}},
            "baz": ["qux"],
        }
        self.assertEqual(expect_dict, new_dict)
        self.assertIs(type(expect_dict), type(new_dict))

    def test_filter_dict_recurse(self):
        """
        Test filtering a dictionary with recursing.
        This will filter out any key-values where the values are falsey or when
        the values *become* falsey after filtering their contents (in case they
        are lists or dicts).
        """
        old_dict = {
            "foo": None,
            "bar": {"baz": {"qux": None, "quux": "", "foo": []}},
            "baz": ["qux"],
            "qux": {},
            "quux": [],
        }
        new_dict = salt.utils.data.filter_falsey(old_dict, recurse_depth=3)
        expect_dict = {"baz": ["qux"]}
        self.assertEqual(expect_dict, new_dict)
        self.assertIs(type(expect_dict), type(new_dict))

    def test_filter_list_no_recurse(self):
        """
        Test filtering a list without recursing.
        This will only filter out items which are falsey.
        """
        old_list = ["foo", None, [], {}, 0, ""]
        new_list = salt.utils.data.filter_falsey(old_list)
        expect_list = ["foo"]
        self.assertEqual(expect_list, new_list)
        self.assertIs(type(expect_list), type(new_list))
        # Ensure nested values are *not* filtered out.
        old_list = [
            "foo",
            ["foo"],
            ["foo", None],
            {"foo": 0},
            {"foo": "bar", "baz": []},
            [{"foo": ""}],
        ]
        new_list = salt.utils.data.filter_falsey(old_list)
        self.assertEqual(old_list, new_list)
        self.assertIs(type(old_list), type(new_list))

    def test_filter_list_recurse(self):
        """
        Test filtering a list with recursing.
        This will filter out any items which are falsey, or which become falsey
        after filtering their contents (in case they are lists or dicts).
        """
        old_list = [
            "foo",
            ["foo"],
            ["foo", None],
            {"foo": 0},
            {"foo": "bar", "baz": []},
            [{"foo": ""}],
        ]
        new_list = salt.utils.data.filter_falsey(old_list, recurse_depth=3)
        expect_list = ["foo", ["foo"], ["foo"], {"foo": "bar"}]
        self.assertEqual(expect_list, new_list)
        self.assertIs(type(expect_list), type(new_list))

    def test_filter_set_no_recurse(self):
        """
        Test filtering a set without recursing.
        Note that a set cannot contain unhashable types, so recursion is not possible.
        """
        old_set = {"foo", None, 0, ""}
        new_set = salt.utils.data.filter_falsey(old_set)
        expect_set = {"foo"}
        self.assertEqual(expect_set, new_set)
        self.assertIs(type(expect_set), type(new_set))

    def test_filter_ordereddict_no_recurse(self):
        """
        Test filtering an OrderedDict without recursing.
        """
        old_dict = OrderedDict(
            [
                ("foo", None),
                (
                    "bar",
                    OrderedDict(
                        [
                            (
                                "baz",
                                OrderedDict([("qux", None), ("quux", ""), ("foo", [])]),
                            )
                        ]
                    ),
                ),
                ("baz", ["qux"]),
                ("qux", {}),
                ("quux", []),
            ]
        )
        new_dict = salt.utils.data.filter_falsey(old_dict)
        expect_dict = OrderedDict(
            [
                (
                    "bar",
                    OrderedDict(
                        [
                            (
                                "baz",
                                OrderedDict([("qux", None), ("quux", ""), ("foo", [])]),
                            )
                        ]
                    ),
                ),
                ("baz", ["qux"]),
            ]
        )
        self.assertEqual(expect_dict, new_dict)
        self.assertIs(type(expect_dict), type(new_dict))

    def test_filter_ordereddict_recurse(self):
        """
        Test filtering an OrderedDict with recursing.
        """
        old_dict = OrderedDict(
            [
                ("foo", None),
                (
                    "bar",
                    OrderedDict(
                        [
                            (
                                "baz",
                                OrderedDict([("qux", None), ("quux", ""), ("foo", [])]),
                            )
                        ]
                    ),
                ),
                ("baz", ["qux"]),
                ("qux", {}),
                ("quux", []),
            ]
        )
        new_dict = salt.utils.data.filter_falsey(old_dict, recurse_depth=3)
        expect_dict = OrderedDict([("baz", ["qux"])])
        self.assertEqual(expect_dict, new_dict)
        self.assertIs(type(expect_dict), type(new_dict))

    def test_filter_list_recurse_limit(self):
        """
        Test filtering a list with recursing, but with a limited depth.
        Note that the top-level is always processed, so a recursion depth of 2
        means that two *additional* levels are processed.
        """
        old_list = [None, [None, [None, [None]]]]
        new_list = salt.utils.data.filter_falsey(old_list, recurse_depth=2)
        self.assertEqual([[[[None]]]], new_list)

    def test_filter_dict_recurse_limit(self):
        """
        Test filtering a dict with recursing, but with a limited depth.
        Note that the top-level is always processed, so a recursion depth of 2
        means that two *additional* levels are processed.
        """
        old_dict = {
            "one": None,
            "foo": {"two": None, "bar": {"three": None, "baz": {"four": None}}},
        }
        new_dict = salt.utils.data.filter_falsey(old_dict, recurse_depth=2)
        self.assertEqual({"foo": {"bar": {"baz": {"four": None}}}}, new_dict)

    def test_filter_exclude_types(self):
        """
        Test filtering a list recursively, but also ignoring (i.e. not filtering)
        out certain types that can be falsey.
        """
        # Ignore int, unicode
        old_list = [
            "foo",
            ["foo"],
            ["foo", None],
            {"foo": 0},
            {"foo": "bar", "baz": []},
            [{"foo": ""}],
        ]
        new_list = salt.utils.data.filter_falsey(
            old_list, recurse_depth=3, ignore_types=[type(0), type("")]
        )
        self.assertEqual(
            ["foo", ["foo"], ["foo"], {"foo": 0}, {"foo": "bar"}, [{"foo": ""}]],
            new_list,
        )
        # Ignore list
        old_list = [
            "foo",
            ["foo"],
            ["foo", None],
            {"foo": 0},
            {"foo": "bar", "baz": []},
            [{"foo": ""}],
        ]
        new_list = salt.utils.data.filter_falsey(
            old_list, recurse_depth=3, ignore_types=[type([])]
        )
        self.assertEqual(
            ["foo", ["foo"], ["foo"], {"foo": "bar", "baz": []}, []], new_list
        )
        # Ignore dict
        old_list = [
            "foo",
            ["foo"],
            ["foo", None],
            {"foo": 0},
            {"foo": "bar", "baz": []},
            [{"foo": ""}],
        ]
        new_list = salt.utils.data.filter_falsey(
            old_list, recurse_depth=3, ignore_types=[type({})]
        )
        self.assertEqual(["foo", ["foo"], ["foo"], {}, {"foo": "bar"}, [{}]], new_list)
        # Ignore NoneType
        old_list = [
            "foo",
            ["foo"],
            ["foo", None],
            {"foo": 0},
            {"foo": "bar", "baz": []},
            [{"foo": ""}],
        ]
        new_list = salt.utils.data.filter_falsey(
            old_list, recurse_depth=3, ignore_types=[type(None)]
        )
        self.assertEqual(["foo", ["foo"], ["foo", None], {"foo": "bar"}], new_list)


class FilterRecursiveDiff(TestCase):
    """
    Test suite for salt.utils.data.recursive_diff
    """

    def test_list_equality(self):
        """
        Test cases where equal lists are compared.
        """
        test_list = [0, 1, 2]
        self.assertEqual({}, salt.utils.data.recursive_diff(test_list, test_list))

        test_list = [[0], [1], [0, 1, 2]]
        self.assertEqual({}, salt.utils.data.recursive_diff(test_list, test_list))

    def test_dict_equality(self):
        """
        Test cases where equal dicts are compared.
        """
        test_dict = {"foo": "bar", "bar": {"baz": {"qux": "quux"}}, "frop": 0}
        self.assertEqual({}, salt.utils.data.recursive_diff(test_dict, test_dict))

    def test_ordereddict_equality(self):
        """
        Test cases where equal OrderedDicts are compared.
        """
        test_dict = OrderedDict(
            [
                ("foo", "bar"),
                ("bar", OrderedDict([("baz", OrderedDict([("qux", "quux")]))])),
                ("frop", 0),
            ]
        )
        self.assertEqual({}, salt.utils.data.recursive_diff(test_dict, test_dict))

    def test_mixed_equality(self):
        """
        Test cases where mixed nested lists and dicts are compared.
        """
        test_data = {
            "foo": "bar",
            "baz": [0, 1, 2],
            "bar": {"baz": [{"qux": "quux"}, {"froop", 0}]},
        }
        self.assertEqual({}, salt.utils.data.recursive_diff(test_data, test_data))

    def test_set_equality(self):
        """
        Test cases where equal sets are compared.
        """
        test_set = {0, 1, 2, 3, "foo"}
        self.assertEqual({}, salt.utils.data.recursive_diff(test_set, test_set))

        # This is a bit of an oddity, as python seems to sort the sets in memory
        # so both sets end up with the same ordering (0..3).
        set_one = {0, 1, 2, 3}
        set_two = {3, 2, 1, 0}
        self.assertEqual({}, salt.utils.data.recursive_diff(set_one, set_two))

    def test_tuple_equality(self):
        """
        Test cases where equal tuples are compared.
        """
        test_tuple = (0, 1, 2, 3, "foo")
        self.assertEqual({}, salt.utils.data.recursive_diff(test_tuple, test_tuple))

    def test_list_inequality(self):
        """
        Test cases where two inequal lists are compared.
        """
        list_one = [0, 1, 2]
        list_two = ["foo", "bar", "baz"]
        expected_result = {"old": list_one, "new": list_two}
        self.assertEqual(
            expected_result, salt.utils.data.recursive_diff(list_one, list_two)
        )
        expected_result = {"new": list_one, "old": list_two}
        self.assertEqual(
            expected_result, salt.utils.data.recursive_diff(list_two, list_one)
        )

        list_one = [0, "foo", 1, "bar"]
        list_two = [1, "foo", 1, "qux"]
        expected_result = {"old": [0, "bar"], "new": [1, "qux"]}
        self.assertEqual(
            expected_result, salt.utils.data.recursive_diff(list_one, list_two)
        )
        expected_result = {"new": [0, "bar"], "old": [1, "qux"]}
        self.assertEqual(
            expected_result, salt.utils.data.recursive_diff(list_two, list_one)
        )

        list_one = [0, 1, [2, 3]]
        list_two = [0, 1, ["foo", "bar"]]
        expected_result = {"old": [[2, 3]], "new": [["foo", "bar"]]}
        self.assertEqual(
            expected_result, salt.utils.data.recursive_diff(list_one, list_two)
        )
        expected_result = {"new": [[2, 3]], "old": [["foo", "bar"]]}
        self.assertEqual(
            expected_result, salt.utils.data.recursive_diff(list_two, list_one)
        )

    def test_dict_inequality(self):
        """
        Test cases where two inequal dicts are compared.
        """
        dict_one = {"foo": 1, "bar": 2, "baz": 3}
        dict_two = {"foo": 2, 1: "bar", "baz": 3}
        expected_result = {"old": {"foo": 1, "bar": 2}, "new": {"foo": 2, 1: "bar"}}
        self.assertEqual(
            expected_result, salt.utils.data.recursive_diff(dict_one, dict_two)
        )
        expected_result = {"new": {"foo": 1, "bar": 2}, "old": {"foo": 2, 1: "bar"}}
        self.assertEqual(
            expected_result, salt.utils.data.recursive_diff(dict_two, dict_one)
        )

        dict_one = {"foo": {"bar": {"baz": 1}}}
        dict_two = {"foo": {"qux": {"baz": 1}}}
        expected_result = {"old": dict_one, "new": dict_two}
        self.assertEqual(
            expected_result, salt.utils.data.recursive_diff(dict_one, dict_two)
        )
        expected_result = {"new": dict_one, "old": dict_two}
        self.assertEqual(
            expected_result, salt.utils.data.recursive_diff(dict_two, dict_one)
        )

    def test_ordereddict_inequality(self):
        """
        Test cases where two inequal OrderedDicts are compared.
        """
        odict_one = OrderedDict([("foo", "bar"), ("bar", "baz")])
        odict_two = OrderedDict([("bar", "baz"), ("foo", "bar")])
        expected_result = {"old": odict_one, "new": odict_two}
        self.assertEqual(
            expected_result, salt.utils.data.recursive_diff(odict_one, odict_two)
        )

    def test_set_inequality(self):
        """
        Test cases where two inequal sets are compared.
        Tricky as the sets are compared zipped, so shuffled sets of equal values
        are considered different.
        """
        set_one = {0, 1, 2, 4}
        set_two = {0, 1, 3, 4}
        expected_result = {"old": {2}, "new": {3}}
        self.assertEqual(
            expected_result, salt.utils.data.recursive_diff(set_one, set_two)
        )
        expected_result = {"new": {2}, "old": {3}}
        self.assertEqual(
            expected_result, salt.utils.data.recursive_diff(set_two, set_one)
        )

        # It is unknown how different python versions will store sets in memory.
        # Python 2.7 seems to sort it (i.e. set_one below becomes {0, 1, 'foo', 'bar'}
        # However Python 3.6.8 stores it differently each run.
        # So just test for "not equal" here.
        set_one = {0, "foo", 1, "bar"}
        set_two = {"foo", 1, "bar", 2}
        expected_result = {}
        self.assertNotEqual(
            expected_result, salt.utils.data.recursive_diff(set_one, set_two)
        )

    def test_mixed_inequality(self):
        """
        Test cases where two mixed dicts/iterables that are different are compared.
        """
        dict_one = {"foo": [1, 2, 3]}
        dict_two = {"foo": [3, 2, 1]}
        expected_result = {"old": {"foo": [1, 3]}, "new": {"foo": [3, 1]}}
        self.assertEqual(
            expected_result, salt.utils.data.recursive_diff(dict_one, dict_two)
        )
        expected_result = {"new": {"foo": [1, 3]}, "old": {"foo": [3, 1]}}
        self.assertEqual(
            expected_result, salt.utils.data.recursive_diff(dict_two, dict_one)
        )

        list_one = [1, 2, {"foo": ["bar", {"foo": 1, "bar": 2}]}]
        list_two = [3, 4, {"foo": ["qux", {"foo": 1, "bar": 2}]}]
        expected_result = {
            "old": [1, 2, {"foo": ["bar"]}],
            "new": [3, 4, {"foo": ["qux"]}],
        }
        self.assertEqual(
            expected_result, salt.utils.data.recursive_diff(list_one, list_two)
        )
        expected_result = {
            "new": [1, 2, {"foo": ["bar"]}],
            "old": [3, 4, {"foo": ["qux"]}],
        }
        self.assertEqual(
            expected_result, salt.utils.data.recursive_diff(list_two, list_one)
        )

        mixed_one = {"foo": {0, 1, 2}, "bar": [0, 1, 2]}
        mixed_two = {"foo": {1, 2, 3}, "bar": [1, 2, 3]}
        expected_result = {
            "old": {"foo": {0}, "bar": [0, 1, 2]},
            "new": {"foo": {3}, "bar": [1, 2, 3]},
        }
        self.assertEqual(
            expected_result, salt.utils.data.recursive_diff(mixed_one, mixed_two)
        )
        expected_result = {
            "new": {"foo": {0}, "bar": [0, 1, 2]},
            "old": {"foo": {3}, "bar": [1, 2, 3]},
        }
        self.assertEqual(
            expected_result, salt.utils.data.recursive_diff(mixed_two, mixed_one)
        )

    def test_tuple_inequality(self):
        """
        Test cases where two tuples that are different are compared.
        """
        tuple_one = (1, 2, 3)
        tuple_two = (3, 2, 1)
        expected_result = {"old": (1, 3), "new": (3, 1)}
        self.assertEqual(
            expected_result, salt.utils.data.recursive_diff(tuple_one, tuple_two)
        )

    def test_list_vs_set(self):
        """
        Test case comparing a list with a set, will be compared unordered.
        """
        mixed_one = [1, 2, 3]
        mixed_two = {3, 2, 1}
        expected_result = {}
        self.assertEqual(
            expected_result, salt.utils.data.recursive_diff(mixed_one, mixed_two)
        )
        self.assertEqual(
            expected_result, salt.utils.data.recursive_diff(mixed_two, mixed_one)
        )

    def test_dict_vs_ordereddict(self):
        """
        Test case comparing a dict with an ordereddict, will be compared unordered.
        """
        test_dict = {"foo": "bar", "bar": "baz"}
        test_odict = OrderedDict([("foo", "bar"), ("bar", "baz")])
        self.assertEqual({}, salt.utils.data.recursive_diff(test_dict, test_odict))
        self.assertEqual({}, salt.utils.data.recursive_diff(test_odict, test_dict))

        test_odict2 = OrderedDict([("bar", "baz"), ("foo", "bar")])
        self.assertEqual({}, salt.utils.data.recursive_diff(test_dict, test_odict2))
        self.assertEqual({}, salt.utils.data.recursive_diff(test_odict2, test_dict))

    def test_list_ignore_ignored(self):
        """
        Test case comparing two lists with ignore-list supplied (which is not used
        when comparing lists).
        """
        list_one = [1, 2, 3]
        list_two = [3, 2, 1]
        expected_result = {"old": [1, 3], "new": [3, 1]}
        self.assertEqual(
            expected_result,
            salt.utils.data.recursive_diff(list_one, list_two, ignore_keys=[1, 3]),
        )

    def test_dict_ignore(self):
        """
        Test case comparing two dicts with ignore-list supplied.
        """
        dict_one = {"foo": 1, "bar": 2, "baz": 3}
        dict_two = {"foo": 3, "bar": 2, "baz": 1}
        expected_result = {"old": {"baz": 3}, "new": {"baz": 1}}
        self.assertEqual(
            expected_result,
            salt.utils.data.recursive_diff(dict_one, dict_two, ignore_keys=["foo"]),
        )

    def test_ordereddict_ignore(self):
        """
        Test case comparing two OrderedDicts with ignore-list supplied.
        """
        odict_one = OrderedDict([("foo", 1), ("bar", 2), ("baz", 3)])
        odict_two = OrderedDict([("baz", 1), ("bar", 2), ("foo", 3)])
        # The key 'foo' will be ignored, which means the key from the other OrderedDict
        # will always be considered "different" since OrderedDicts are compared ordered.
        expected_result = {
            "old": OrderedDict([("baz", 3)]),
            "new": OrderedDict([("baz", 1)]),
        }
        self.assertEqual(
            expected_result,
            salt.utils.data.recursive_diff(odict_one, odict_two, ignore_keys=["foo"]),
        )

    def test_dict_vs_ordereddict_ignore(self):
        """
        Test case comparing a dict with an OrderedDict with ignore-list supplied.
        """
        dict_one = {"foo": 1, "bar": 2, "baz": 3}
        odict_two = OrderedDict([("foo", 3), ("bar", 2), ("baz", 1)])
        expected_result = {"old": {"baz": 3}, "new": OrderedDict([("baz", 1)])}
        self.assertEqual(
            expected_result,
            salt.utils.data.recursive_diff(dict_one, odict_two, ignore_keys=["foo"]),
        )

    def test_mixed_nested_ignore(self):
        """
        Test case comparing mixed, nested items with ignore-list supplied.
        """
        dict_one = {"foo": [1], "bar": {"foo": 1, "bar": 2}, "baz": 3}
        dict_two = {"foo": [2], "bar": {"foo": 3, "bar": 2}, "baz": 1}
        expected_result = {"old": {"baz": 3}, "new": {"baz": 1}}
        self.assertEqual(
            expected_result,
            salt.utils.data.recursive_diff(dict_one, dict_two, ignore_keys=["foo"]),
        )

    def test_ordered_dict_unequal_length(self):
        """
        Test case comparing two OrderedDicts of unequal length.
        """
        odict_one = OrderedDict([("foo", 1), ("bar", 2), ("baz", 3)])
        odict_two = OrderedDict([("foo", 1), ("bar", 2)])
        expected_result = {"old": OrderedDict([("baz", 3)]), "new": {}}
        self.assertEqual(
            expected_result, salt.utils.data.recursive_diff(odict_one, odict_two)
        )

    def test_list_unequal_length(self):
        """
        Test case comparing two lists of unequal length.
        """
        list_one = [1, 2, 3]
        list_two = [1, 2, 3, 4]
        expected_result = {"old": [], "new": [4]}
        self.assertEqual(
            expected_result, salt.utils.data.recursive_diff(list_one, list_two)
        )

    def test_set_unequal_length(self):
        """
        Test case comparing two sets of unequal length.
        This does not do anything special, as it is unordered.
        """
        set_one = {1, 2, 3}
        set_two = {4, 3, 2, 1}
        expected_result = {"old": set(), "new": {4}}
        self.assertEqual(
            expected_result, salt.utils.data.recursive_diff(set_one, set_two)
        )

    def test_tuple_unequal_length(self):
        """
        Test case comparing two tuples of unequal length.
        This should be the same as comparing two ordered lists.
        """
        tuple_one = (1, 2, 3)
        tuple_two = (1, 2, 3, 4)
        expected_result = {"old": (), "new": (4,)}
        self.assertEqual(
            expected_result, salt.utils.data.recursive_diff(tuple_one, tuple_two)
        )

    def test_list_unordered(self):
        """
        Test case comparing two lists unordered.
        """
        list_one = [1, 2, 3, 4]
        list_two = [4, 3, 2]
        expected_result = {"old": [1], "new": []}
        self.assertEqual(
            expected_result,
            salt.utils.data.recursive_diff(list_one, list_two, ignore_order=True),
        )

    def test_mixed_nested_unordered(self):
        """
        Test case comparing nested dicts/lists unordered.
        """
        dict_one = {"foo": {"bar": [1, 2, 3]}, "bar": [{"foo": 4}, 0]}
        dict_two = {"foo": {"bar": [3, 2, 1]}, "bar": [0, {"foo": 4}]}
        expected_result = {}
        self.assertEqual(
            expected_result,
            salt.utils.data.recursive_diff(dict_one, dict_two, ignore_order=True),
        )
        expected_result = {
            "old": {"foo": {"bar": [1, 3]}, "bar": [{"foo": 4}, 0]},
            "new": {"foo": {"bar": [3, 1]}, "bar": [0, {"foo": 4}]},
        }
        self.assertEqual(
            expected_result, salt.utils.data.recursive_diff(dict_one, dict_two)
        )

    def test_ordered_dict_unordered(self):
        """
        Test case comparing OrderedDicts unordered.
        """
        odict_one = OrderedDict([("foo", 1), ("bar", 2), ("baz", 3)])
        odict_two = OrderedDict([("baz", 3), ("bar", 2), ("foo", 1)])
        expected_result = {}
        self.assertEqual(
            expected_result,
            salt.utils.data.recursive_diff(odict_one, odict_two, ignore_order=True),
        )

    def test_ignore_missing_keys_dict(self):
        """
        Test case ignoring missing keys on a comparison of dicts.
        """
        dict_one = {"foo": 1, "bar": 2, "baz": 3}
        dict_two = {"bar": 3}
        expected_result = {"old": {"bar": 2}, "new": {"bar": 3}}
        self.assertEqual(
            expected_result,
            salt.utils.data.recursive_diff(
                dict_one, dict_two, ignore_missing_keys=True
            ),
        )

    def test_ignore_missing_keys_ordered_dict(self):
        """
        Test case not ignoring missing keys on a comparison of OrderedDicts.
        """
        odict_one = OrderedDict([("foo", 1), ("bar", 2), ("baz", 3)])
        odict_two = OrderedDict([("bar", 3)])
        expected_result = {"old": odict_one, "new": odict_two}
        self.assertEqual(
            expected_result,
            salt.utils.data.recursive_diff(
                odict_one, odict_two, ignore_missing_keys=True
            ),
        )

    def test_ignore_missing_keys_recursive(self):
        """
        Test case ignoring missing keys on a comparison of nested dicts.
        """
        dict_one = {"foo": {"bar": 2, "baz": 3}}
        dict_two = {"foo": {"baz": 3}}
        expected_result = {}
        self.assertEqual(
            expected_result,
            salt.utils.data.recursive_diff(
                dict_one, dict_two, ignore_missing_keys=True
            ),
        )
        # Compare from dict-in-dict
        dict_two = {}
        self.assertEqual(
            expected_result,
            salt.utils.data.recursive_diff(
                dict_one, dict_two, ignore_missing_keys=True
            ),
        )
        # Compare from dict-in-list
        dict_one = {"foo": ["bar", {"baz": 3}]}
        dict_two = {"foo": ["bar", {}]}
        self.assertEqual(
            expected_result,
            salt.utils.data.recursive_diff(
                dict_one, dict_two, ignore_missing_keys=True
            ),
        )



import pytest
from saltfactories.utils import random_string


def transport_ids(value):
    return "Transport({})".format(value)


@pytest.fixture(params=("zeromq", "tcp"), ids=transport_ids)
def transport(request):
    return request.param


@pytest.fixture
def salt_master(salt_factories, transport):
    config_defaults = {
        "transport": transport,
        "auto_accept": True,
        "sign_pub_messages": False,
    }
    factory = salt_factories.salt_master_daemon(
        random_string("server-{}-master-".format(transport)),
        defaults=config_defaults,
    )
    return factory


@pytest.fixture
def salt_minion(salt_master, transport):
    config_defaults = {
        "transport": transport,
        "master_ip": "127.0.0.1",
        "master_port": salt_master.config["ret_port"],
        "auth_timeout": 5,
        "auth_tries": 1,
        "master_uri": "tcp://127.0.0.1:{}".format(salt_master.config["ret_port"]),
    }
    factory = salt_master.salt_minion_daemon(
        random_string("server-{}-minion-".format(transport)),
        defaults=config_defaults,
    )
    return factory



"""
Module for handling openstack keystone calls.

:optdepends:    - keystoneclient Python adapter
:configuration: This module is not usable until the following are specified
    either in a pillar or in the minion's config file:

    .. code-block:: yaml

        keystone.user: admin
        keystone.password: verybadpass
        keystone.tenant: admin
        keystone.tenant_id: f80919baedab48ec8931f200c65a50df
        keystone.auth_url: 'http://127.0.0.1:5000/v2.0/'
        keystone.verify_ssl: True

    OR (for token based authentication)

    .. code-block:: yaml

        keystone.token: 'ADMIN'
        keystone.endpoint: 'http://127.0.0.1:35357/v2.0'

    If configuration for multiple openstack accounts is required, they can be
    set up as different configuration profiles. For example:

    .. code-block:: yaml

        openstack1:
          keystone.user: admin
          keystone.password: verybadpass
          keystone.tenant: admin
          keystone.tenant_id: f80919baedab48ec8931f200c65a50df
          keystone.auth_url: 'http://127.0.0.1:5000/v2.0/'
          keystone.verify_ssl: True

        openstack2:
          keystone.user: admin
          keystone.password: verybadpass
          keystone.tenant: admin
          keystone.tenant_id: f80919baedab48ec8931f200c65a50df
          keystone.auth_url: 'http://127.0.0.2:5000/v2.0/'
          keystone.verify_ssl: True

    With this configuration in place, any of the keystone functions can make use
    of a configuration profile by declaring it explicitly.
    For example:

    .. code-block:: bash

        salt '*' keystone.tenant_list profile=openstack1
"""

import logging

import salt.utils.http

HAS_KEYSTONE = False
try:
    # pylint: disable=import-error
    from keystoneclient.v2_0 import client
    import keystoneclient.exceptions

    HAS_KEYSTONE = True
    from keystoneclient.v3 import client as client3
    from keystoneclient import discover
    from keystoneauth1 import session
    from keystoneauth1.identity import generic

    # pylint: enable=import-error
except ImportError:
    pass

_OS_IDENTITY_API_VERSION = 2
_TENANTS = "tenants"

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load this module if keystone
    is installed on this minion.
    """
    if HAS_KEYSTONE:
        return "keystone"
    return (
        False,
        "keystone execution module cannot be loaded: keystoneclient python library not"
        " available.",
    )


__opts__ = {}


def _get_kwargs(profile=None, **connection_args):
    """
    get connection args
    """
    if profile:
        prefix = profile + ":keystone."
    else:
        prefix = "keystone."

    def get(key, default=None):
        """
        look in connection_args first, then default to config file
        """
        return connection_args.get(
            "connection_" + key, __salt__["config.get"](prefix + key, default)
        )

    user = get("user", "admin")
    password = get("password", "ADMIN")
    tenant = get("tenant", "admin")
    tenant_id = get("tenant_id")
    auth_url = get("auth_url", "http://127.0.0.1:35357/v2.0/")
    insecure = get("insecure", False)
    token = get("token")
    endpoint = get("endpoint", "http://127.0.0.1:35357/v2.0")
    user_domain_name = get("user_domain_name", "Default")
    project_domain_name = get("project_domain_name", "Default")
    verify_ssl = get("verify_ssl", True)
    if token:
        kwargs = {"token": token, "endpoint": endpoint}
    else:
        kwargs = {
            "username": user,
            "password": password,
            "tenant_name": tenant,
            "tenant_id": tenant_id,
            "auth_url": auth_url,
            "user_domain_name": user_domain_name,
            "project_domain_name": project_domain_name,
        }
        # 'insecure' keyword not supported by all v2.0 keystone clients
        #   this ensures it's only passed in when defined
        if insecure:
            kwargs["insecure"] = True
    kwargs["verify_ssl"] = verify_ssl
    return kwargs


def api_version(profile=None, **connection_args):
    """
    Returns the API version derived from endpoint's response.

    CLI Example:

    .. code-block:: bash

        salt '*' keystone.api_version
    """
    kwargs = _get_kwargs(profile=profile, **connection_args)
    auth_url = kwargs.get("auth_url", kwargs.get("endpoint", None))
    try:
        return salt.utils.http.query(
            auth_url, decode=True, decode_type="json", verify_ssl=kwargs["verify_ssl"]
        )["dict"]["version"]["id"]
    except KeyError:
        return None


def auth(profile=None, **connection_args):
    """
    Set up keystone credentials. Only intended to be used within Keystone-enabled modules.

    CLI Example:

    .. code-block:: bash

        salt '*' keystone.auth
    """
    __utils__["versions.warn_until"](
        "Phosphorus",
        "The keystone module has been deprecated and will be removed in {version}.  "
        "Please update to using the keystoneng module",
    )
    kwargs = _get_kwargs(profile=profile, **connection_args)

    disc = discover.Discover(auth_url=kwargs["auth_url"])
    v2_auth_url = disc.url_for("v2.0")
    v3_auth_url = disc.url_for("v3.0")
    if v3_auth_url:
        global _OS_IDENTITY_API_VERSION
        global _TENANTS
        _OS_IDENTITY_API_VERSION = 3
        _TENANTS = "projects"
        kwargs["auth_url"] = v3_auth_url
    else:
        kwargs["auth_url"] = v2_auth_url
        kwargs.pop("user_domain_name")
        kwargs.pop("project_domain_name")
    auth = generic.Password(**kwargs)
    sess = session.Session(auth=auth)
    ks_cl = disc.create_client(session=sess)
    return ks_cl


def ec2_credentials_create(
    user_id=None,
    name=None,
    tenant_id=None,
    tenant=None,
    profile=None,
    **connection_args
):
    """
    Create EC2-compatible credentials for user per tenant

    CLI Examples:

    .. code-block:: bash

        salt '*' keystone.ec2_credentials_create name=admin tenant=admin

        salt '*' keystone.ec2_credentials_create \
        user_id=c965f79c4f864eaaa9c3b41904e67082 \
        tenant_id=722787eb540849158668370dc627ec5f
    """
    kstone = auth(profile, **connection_args)

    if name:
        user_id = user_get(name=name, profile=profile, **connection_args)[name]["id"]
    if not user_id:
        return {"Error": "Could not resolve User ID"}

    if tenant:
        tenant_id = tenant_get(name=tenant, profile=profile, **connection_args)[tenant][
            "id"
        ]
    if not tenant_id:
        return {"Error": "Could not resolve Tenant ID"}

    newec2 = kstone.ec2.create(user_id, tenant_id)
    return {
        "access": newec2.access,
        "secret": newec2.secret,
        "tenant_id": newec2.tenant_id,
        "user_id": newec2.user_id,
    }


def ec2_credentials_delete(
    user_id=None, name=None, access_key=None, profile=None, **connection_args
):
    """
    Delete EC2-compatible credentials

    CLI Examples:

    .. code-block:: bash

        salt '*' keystone.ec2_credentials_delete \
        860f8c2c38ca4fab989f9bc56a061a64 access_key=5f66d2f24f604b8bb9cd28886106f442

        salt '*' keystone.ec2_credentials_delete name=admin \
        access_key=5f66d2f24f604b8bb9cd28886106f442
    """
    kstone = auth(profile, **connection_args)

    if name:
        user_id = user_get(name=name, profile=None, **connection_args)[name]["id"]
    if not user_id:
        return {"Error": "Could not resolve User ID"}
    kstone.ec2.delete(user_id, access_key)
    return 'ec2 key "{}" deleted under user id "{}"'.format(access_key, user_id)


def ec2_credentials_get(
    user_id=None, name=None, access=None, profile=None, **connection_args
):
    """
    Return ec2_credentials for a user (keystone ec2-credentials-get)

    CLI Examples:

    .. code-block:: bash

        salt '*' keystone.ec2_credentials_get c965f79c4f864eaaa9c3b41904e67082 access=722787eb540849158668370
        salt '*' keystone.ec2_credentials_get user_id=c965f79c4f864eaaa9c3b41904e67082 access=722787eb540849158668370
        salt '*' keystone.ec2_credentials_get name=nova access=722787eb540849158668370dc627ec5f
    """
    kstone = auth(profile, **connection_args)
    ret = {}
    if name:
        for user in kstone.users.list():
            if user.name == name:
                user_id = user.id
                break
    if not user_id:
        return {"Error": "Unable to resolve user id"}
    if not access:
        return {"Error": "Access key is required"}
    ec2_credentials = kstone.ec2.get(
        user_id=user_id, access=access, profile=profile, **connection_args
    )
    ret[ec2_credentials.user_id] = {
        "user_id": ec2_credentials.user_id,
        "tenant": ec2_credentials.tenant_id,
        "access": ec2_credentials.access,
        "secret": ec2_credentials.secret,
    }
    return ret


def ec2_credentials_list(user_id=None, name=None, profile=None, **connection_args):
    """
    Return a list of ec2_credentials for a specific user (keystone ec2-credentials-list)

    CLI Examples:

    .. code-block:: bash

        salt '*' keystone.ec2_credentials_list 298ce377245c4ec9b70e1c639c89e654
        salt '*' keystone.ec2_credentials_list user_id=298ce377245c4ec9b70e1c639c89e654
        salt '*' keystone.ec2_credentials_list name=jack
    """
    kstone = auth(profile, **connection_args)
    ret = {}
    if name:
        for user in kstone.users.list():
            if user.name == name:
                user_id = user.id
                break
    if not user_id:
        return {"Error": "Unable to resolve user id"}
    for ec2_credential in kstone.ec2.list(user_id):
        ret[ec2_credential.user_id] = {
            "user_id": ec2_credential.user_id,
            "tenant_id": ec2_credential.tenant_id,
            "access": ec2_credential.access,
            "secret": ec2_credential.secret,
        }
    return ret


def endpoint_get(service, region=None, profile=None, interface=None, **connection_args):
    """
    Return a specific endpoint (keystone endpoint-get)

    CLI Example:

    .. code-block:: bash

        salt 'v2' keystone.endpoint_get nova [region=RegionOne]

        salt 'v3' keystone.endpoint_get nova interface=admin [region=RegionOne]
    """
    auth(profile, **connection_args)
    services = service_list(profile, **connection_args)
    if service not in services:
        return {"Error": "Could not find the specified service"}
    service_id = services[service]["id"]
    endpoints = endpoint_list(profile, **connection_args)

    e = [
        _f
        for _f in [
            e
            if e["service_id"] == service_id
            and (e["region"] == region if region else True)
            and (e["interface"] == interface if interface else True)
            else None
            for e in endpoints.values()
        ]
        if _f
    ]
    if len(e) > 1:
        return {
            "Error": (
                "Multiple endpoints found ({}) for the {} service. Please specify"
                " region.".format(e, service)
            )
        }
    if len(e) == 1:
        return e[0]
    return {"Error": "Could not find endpoint for the specified service"}


def endpoint_list(profile=None, **connection_args):
    """
    Return a list of available endpoints (keystone endpoints-list)

    CLI Example:

    .. code-block:: bash

        salt '*' keystone.endpoint_list
    """
    kstone = auth(profile, **connection_args)
    ret = {}

    for endpoint in kstone.endpoints.list():
        ret[endpoint.id] = {
            value: getattr(endpoint, value)
            for value in dir(endpoint)
            if not value.startswith("_")
            and isinstance(getattr(endpoint, value), (str, dict, bool))
        }
    return ret


def endpoint_create(
    service,
    publicurl=None,
    internalurl=None,
    adminurl=None,
    region=None,
    profile=None,
    url=None,
    interface=None,
    **connection_args
):
    """
    Create an endpoint for an Openstack service

    CLI Examples:

    .. code-block:: bash

        salt 'v2' keystone.endpoint_create nova 'http://public/url' 'http://internal/url' 'http://adminurl/url' region

        salt 'v3' keystone.endpoint_create nova url='http://public/url' interface='public' region='RegionOne'
    """
    kstone = auth(profile, **connection_args)
    keystone_service = service_get(name=service, profile=profile, **connection_args)
    if not keystone_service or "Error" in keystone_service:
        return {"Error": "Could not find the specified service"}

    if _OS_IDENTITY_API_VERSION > 2:
        kstone.endpoints.create(
            service=keystone_service[service]["id"],
            region_id=region,
            url=url,
            interface=interface,
        )
    else:
        kstone.endpoints.create(
            region=region,
            service_id=keystone_service[service]["id"],
            publicurl=publicurl,
            adminurl=adminurl,
            internalurl=internalurl,
        )
    return endpoint_get(service, region, profile, interface, **connection_args)


def endpoint_delete(
    service, region=None, profile=None, interface=None, **connection_args
):
    """
    Delete endpoints of an Openstack service

    CLI Examples:

    .. code-block:: bash

        salt 'v2' keystone.endpoint_delete nova [region=RegionOne]

        salt 'v3' keystone.endpoint_delete nova interface=admin [region=RegionOne]
    """
    kstone = auth(profile, **connection_args)
    endpoint = endpoint_get(service, region, profile, interface, **connection_args)
    if not endpoint or "Error" in endpoint:
        return {"Error": "Could not find any endpoints for the service"}
    kstone.endpoints.delete(endpoint["id"])
    endpoint = endpoint_get(service, region, profile, interface, **connection_args)
    if not endpoint or "Error" in endpoint:
        return True


def role_create(name, profile=None, **connection_args):
    """
    Create a named role.

    CLI Example:

    .. code-block:: bash

        salt '*' keystone.role_create admin
    """

    kstone = auth(profile, **connection_args)
    if "Error" not in role_get(name=name, profile=profile, **connection_args):
        return {"Error": 'Role "{}" already exists'.format(name)}
    kstone.roles.create(name)
    return role_get(name=name, profile=profile, **connection_args)


def role_delete(role_id=None, name=None, profile=None, **connection_args):
    """
    Delete a role (keystone role-delete)

    CLI Examples:

    .. code-block:: bash

        salt '*' keystone.role_delete c965f79c4f864eaaa9c3b41904e67082
        salt '*' keystone.role_delete role_id=c965f79c4f864eaaa9c3b41904e67082
        salt '*' keystone.role_delete name=admin
    """
    kstone = auth(profile, **connection_args)

    if name:
        for role in kstone.roles.list():
            if role.name == name:
                role_id = role.id
                break
    if not role_id:
        return {"Error": "Unable to resolve role id"}

    role = kstone.roles.get(role_id)
    kstone.roles.delete(role)

    ret = "Role ID {} deleted".format(role_id)
    if name:
        ret += " ({})".format(name)
    return ret


def role_get(role_id=None, name=None, profile=None, **connection_args):
    """
    Return a specific roles (keystone role-get)

    CLI Examples:

    .. code-block:: bash

        salt '*' keystone.role_get c965f79c4f864eaaa9c3b41904e67082
        salt '*' keystone.role_get role_id=c965f79c4f864eaaa9c3b41904e67082
        salt '*' keystone.role_get name=nova
    """
    kstone = auth(profile, **connection_args)
    ret = {}
    if name:
        for role in kstone.roles.list():
            if role.name == name:
                role_id = role.id
                break
    if not role_id:
        return {"Error": "Unable to resolve role id"}
    role = kstone.roles.get(role_id)

    ret[role.name] = {"id": role.id, "name": role.name}
    return ret


def role_list(profile=None, **connection_args):
    """
    Return a list of available roles (keystone role-list)

    CLI Example:

    .. code-block:: bash

        salt '*' keystone.role_list
    """
    kstone = auth(profile, **connection_args)
    ret = {}
    for role in kstone.roles.list():
        ret[role.name] = {
            value: getattr(role, value)
            for value in dir(role)
            if not value.startswith("_")
            and isinstance(getattr(role, value), (str, dict, bool))
        }
    return ret


def service_create(
    name, service_type, description=None, profile=None, **connection_args
):
    """
    Add service to Keystone service catalog

    CLI Examples:

    .. code-block:: bash

        salt '*' keystone.service_create nova compute \
'OpenStack Compute Service'
    """
    kstone = auth(profile, **connection_args)
    service = kstone.services.create(name, service_type, description=description)
    return service_get(service.id, profile=profile, **connection_args)


def service_delete(service_id=None, name=None, profile=None, **connection_args):
    """
    Delete a service from Keystone service catalog

    CLI Examples:

    .. code-block:: bash

        salt '*' keystone.service_delete c965f79c4f864eaaa9c3b41904e67082
        salt '*' keystone.service_delete name=nova
    """
    kstone = auth(profile, **connection_args)
    if name:
        service_id = service_get(name=name, profile=profile, **connection_args)[name][
            "id"
        ]
    kstone.services.delete(service_id)
    return 'Keystone service ID "{}" deleted'.format(service_id)


def service_get(service_id=None, name=None, profile=None, **connection_args):
    """
    Return a specific services (keystone service-get)

    CLI Examples:

    .. code-block:: bash

        salt '*' keystone.service_get c965f79c4f864eaaa9c3b41904e67082
        salt '*' keystone.service_get service_id=c965f79c4f864eaaa9c3b41904e67082
        salt '*' keystone.service_get name=nova
    """
    kstone = auth(profile, **connection_args)
    ret = {}
    if name:
        for service in kstone.services.list():
            if service.name == name:
                service_id = service.id
                break
    if not service_id:
        return {"Error": "Unable to resolve service id"}
    service = kstone.services.get(service_id)
    ret[service.name] = {
        value: getattr(service, value)
        for value in dir(service)
        if not value.startswith("_")
        and isinstance(getattr(service, value), (str, dict, bool))
    }
    return ret


def service_list(profile=None, **connection_args):
    """
    Return a list of available services (keystone services-list)

    CLI Example:

    .. code-block:: bash

        salt '*' keystone.service_list
    """
    kstone = auth(profile, **connection_args)
    ret = {}
    for service in kstone.services.list():
        ret[service.name] = {
            value: getattr(service, value)
            for value in dir(service)
            if not value.startswith("_")
            and isinstance(getattr(service, value), (str, dict, bool))
        }
    return ret


def tenant_create(
    name, description=None, enabled=True, profile=None, **connection_args
):
    """
    Create a keystone tenant

    CLI Examples:

    .. code-block:: bash

        salt '*' keystone.tenant_create nova description='nova tenant'
        salt '*' keystone.tenant_create test enabled=False
    """
    kstone = auth(profile, **connection_args)
    new = getattr(kstone, _TENANTS, None).create(name, description, enabled)
    return tenant_get(new.id, profile=profile, **connection_args)


def project_create(
    name, domain, description=None, enabled=True, profile=None, **connection_args
):
    """
    Create a keystone project.
    Overrides keystone tenant_create form api V2. For keystone api V3.

    .. versionadded:: 2016.11.0

    name
        The project name, which must be unique within the owning domain.

    domain
        The domain name.

    description
        The project description.

    enabled
        Enables or disables the project.

    profile
        Configuration profile - if configuration for multiple openstack accounts required.

    CLI Examples:

    .. code-block:: bash

        salt '*' keystone.project_create nova default description='Nova Compute Project'
        salt '*' keystone.project_create test default enabled=False
    """
    kstone = auth(profile, **connection_args)
    new = getattr(kstone, _TENANTS, None).create(
        name=name, domain=domain, description=description, enabled=enabled
    )
    return tenant_get(new.id, profile=profile, **connection_args)


def tenant_delete(tenant_id=None, name=None, profile=None, **connection_args):
    """
    Delete a tenant (keystone tenant-delete)

    CLI Examples:

    .. code-block:: bash

        salt '*' keystone.tenant_delete c965f79c4f864eaaa9c3b41904e67082
        salt '*' keystone.tenant_delete tenant_id=c965f79c4f864eaaa9c3b41904e67082
        salt '*' keystone.tenant_delete name=demo
    """
    kstone = auth(profile, **connection_args)
    if name:
        for tenant in getattr(kstone, _TENANTS, None).list():
            if tenant.name == name:
                tenant_id = tenant.id
                break
    if not tenant_id:
        return {"Error": "Unable to resolve tenant id"}
    getattr(kstone, _TENANTS, None).delete(tenant_id)
    ret = "Tenant ID {} deleted".format(tenant_id)
    if name:

        ret += " ({})".format(name)
    return ret


def project_delete(project_id=None, name=None, profile=None, **connection_args):
    """
    Delete a project (keystone project-delete).
    Overrides keystone tenant-delete form api V2. For keystone api V3 only.

    .. versionadded:: 2016.11.0

    project_id
        The project id.

    name
        The project name.

    profile
        Configuration profile - if configuration for multiple openstack accounts required.

    CLI Examples:

    .. code-block:: bash

        salt '*' keystone.project_delete c965f79c4f864eaaa9c3b41904e67082
        salt '*' keystone.project_delete project_id=c965f79c4f864eaaa9c3b41904e67082
        salt '*' keystone.project_delete name=demo
    """
    auth(profile, **connection_args)

    if _OS_IDENTITY_API_VERSION > 2:
        return tenant_delete(
            tenant_id=project_id, name=name, profile=None, **connection_args
        )
    else:
        return False


def tenant_get(tenant_id=None, name=None, profile=None, **connection_args):
    """
    Return a specific tenants (keystone tenant-get)

    CLI Examples:

    .. code-block:: bash

        salt '*' keystone.tenant_get c965f79c4f864eaaa9c3b41904e67082
        salt '*' keystone.tenant_get tenant_id=c965f79c4f864eaaa9c3b41904e67082
        salt '*' keystone.tenant_get name=nova
    """
    kstone = auth(profile, **connection_args)
    ret = {}

    if name:
        for tenant in getattr(kstone, _TENANTS, None).list():
            if tenant.name == name:
                tenant_id = tenant.id
                break
    if not tenant_id:
        return {"Error": "Unable to resolve tenant id"}
    tenant = getattr(kstone, _TENANTS, None).get(tenant_id)
    ret[tenant.name] = {
        value: getattr(tenant, value)
        for value in dir(tenant)
        if not value.startswith("_")
        and isinstance(getattr(tenant, value), (str, dict, bool))
    }
    return ret


def project_get(project_id=None, name=None, profile=None, **connection_args):
    """
    Return a specific projects (keystone project-get)
    Overrides keystone tenant-get form api V2.
    For keystone api V3 only.

    .. versionadded:: 2016.11.0

    project_id
        The project id.

    name
        The project name.

    profile
        Configuration profile - if configuration for multiple openstack accounts required.

    CLI Examples:

    .. code-block:: bash

        salt '*' keystone.project_get c965f79c4f864eaaa9c3b41904e67082
        salt '*' keystone.project_get project_id=c965f79c4f864eaaa9c3b41904e67082
        salt '*' keystone.project_get name=nova
    """
    auth(profile, **connection_args)

    if _OS_IDENTITY_API_VERSION > 2:
        return tenant_get(
            tenant_id=project_id, name=name, profile=None, **connection_args
        )
    else:
        return False


def tenant_list(profile=None, **connection_args):
    """
    Return a list of available tenants (keystone tenants-list)

    CLI Example:

    .. code-block:: bash

        salt '*' keystone.tenant_list
    """
    kstone = auth(profile, **connection_args)
    ret = {}

    for tenant in getattr(kstone, _TENANTS, None).list():
        ret[tenant.name] = {
            value: getattr(tenant, value)
            for value in dir(tenant)
            if not value.startswith("_")
            and isinstance(getattr(tenant, value), (str, dict, bool))
        }
    return ret


def project_list(profile=None, **connection_args):
    """
    Return a list of available projects (keystone projects-list).
    Overrides keystone tenants-list form api V2.
    For keystone api V3 only.

    .. versionadded:: 2016.11.0

    profile
        Configuration profile - if configuration for multiple openstack accounts required.

    CLI Example:

    .. code-block:: bash

        salt '*' keystone.project_list
    """
    auth(profile, **connection_args)

    if _OS_IDENTITY_API_VERSION > 2:
        return tenant_list(profile, **connection_args)
    else:
        return False


def tenant_update(
    tenant_id=None,
    name=None,
    description=None,
    enabled=None,
    profile=None,
    **connection_args
):
    """
    Update a tenant's information (keystone tenant-update)
    The following fields may be updated: name, description, enabled.
    Can only update name if targeting by ID

    CLI Examples:

    .. code-block:: bash

        salt '*' keystone.tenant_update name=admin enabled=True
        salt '*' keystone.tenant_update c965f79c4f864eaaa9c3b41904e67082 name=admin email=admin@domain.com
    """
    kstone = auth(profile, **connection_args)

    if not tenant_id:
        for tenant in getattr(kstone, _TENANTS, None).list():
            if tenant.name == name:
                tenant_id = tenant.id
                break
    if not tenant_id:
        return {"Error": "Unable to resolve tenant id"}

    tenant = getattr(kstone, _TENANTS, None).get(tenant_id)
    if not name:
        name = tenant.name
    if not description:
        description = tenant.description
    if enabled is None:
        enabled = tenant.enabled
    updated = getattr(kstone, _TENANTS, None).update(
        tenant_id, name=name, description=description, enabled=enabled
    )

    return {
        value: getattr(updated, value)
        for value in dir(updated)
        if not value.startswith("_")
        and isinstance(getattr(updated, value), (str, dict, bool))
    }


def project_update(
    project_id=None,
    name=None,
    description=None,
    enabled=None,
    profile=None,
    **connection_args
):
    """
    Update a tenant's information (keystone project-update)
    The following fields may be updated: name, description, enabled.
    Can only update name if targeting by ID

    Overrides keystone tenant_update form api V2.
    For keystone api V3 only.

    .. versionadded:: 2016.11.0

    project_id
        The project id.

    name
        The project name, which must be unique within the owning domain.

    description
        The project description.

    enabled
        Enables or disables the project.

    profile
        Configuration profile - if configuration for multiple openstack accounts required.

    CLI Examples:

    .. code-block:: bash

        salt '*' keystone.project_update name=admin enabled=True
        salt '*' keystone.project_update c965f79c4f864eaaa9c3b41904e67082 name=admin email=admin@domain.com
    """
    auth(profile, **connection_args)

    if _OS_IDENTITY_API_VERSION > 2:
        return tenant_update(
            tenant_id=project_id,
            name=name,
            description=description,
            enabled=enabled,
            profile=profile,
            **connection_args
        )
    else:
        return False


def token_get(profile=None, **connection_args):
    """
    Return the configured tokens (keystone token-get)

    CLI Example:

    .. code-block:: bash

        salt '*' keystone.token_get c965f79c4f864eaaa9c3b41904e67082
    """
    kstone = auth(profile, **connection_args)
    token = kstone.service_catalog.get_token()
    return {
        "id": token["id"],
        "expires": token["expires"],
        "user_id": token["user_id"],
        "tenant_id": token["tenant_id"],
    }


def user_list(profile=None, **connection_args):
    """
    Return a list of available users (keystone user-list)

    CLI Example:

    .. code-block:: bash

        salt '*' keystone.user_list
    """
    kstone = auth(profile, **connection_args)
    ret = {}
    for user in kstone.users.list():
        ret[user.name] = {
            value: getattr(user, value, None)
            for value in dir(user)
            if not value.startswith("_")
            and isinstance(getattr(user, value, None), (str, dict, bool))
        }
        tenant_id = getattr(user, "tenantId", None)
        if tenant_id:
            ret[user.name]["tenant_id"] = tenant_id
    return ret


def user_get(user_id=None, name=None, profile=None, **connection_args):
    """
    Return a specific users (keystone user-get)

    CLI Examples:

    .. code-block:: bash

        salt '*' keystone.user_get c965f79c4f864eaaa9c3b41904e67082
        salt '*' keystone.user_get user_id=c965f79c4f864eaaa9c3b41904e67082
        salt '*' keystone.user_get name=nova
    """
    kstone = auth(profile, **connection_args)
    ret = {}
    if name:
        for user in kstone.users.list():
            if user.name == name:
                user_id = user.id
                break
    if not user_id:
        return {"Error": "Unable to resolve user id"}
    try:
        user = kstone.users.get(user_id)
    except keystoneclient.exceptions.NotFound:
        msg = "Could not find user '{}'".format(user_id)
        log.error(msg)
        return {"Error": msg}

    ret[user.name] = {
        value: getattr(user, value, None)
        for value in dir(user)
        if not value.startswith("_")
        and isinstance(getattr(user, value, None), (str, dict, bool))
    }

    tenant_id = getattr(user, "tenantId", None)
    if tenant_id:
        ret[user.name]["tenant_id"] = tenant_id
    return ret


def user_create(
    name,
    password,
    email,
    tenant_id=None,
    enabled=True,
    profile=None,
    project_id=None,
    description=None,
    **connection_args
):
    """
    Create a user (keystone user-create)

    CLI Examples:

    .. code-block:: bash

        salt '*' keystone.user_create name=jack password=zero email=jack@halloweentown.org \
        tenant_id=a28a7b5a999a455f84b1f5210264375e enabled=True
    """
    kstone = auth(profile, **connection_args)

    if _OS_IDENTITY_API_VERSION > 2:
        if tenant_id and not project_id:
            project_id = tenant_id
        item = kstone.users.create(
            name=name,
            password=password,
            email=email,
            project_id=project_id,
            enabled=enabled,
            description=description,
        )
    else:
        item = kstone.users.create(
            name=name,
            password=password,
            email=email,
            tenant_id=tenant_id,
            enabled=enabled,
        )
    return user_get(item.id, profile=profile, **connection_args)


def user_delete(user_id=None, name=None, profile=None, **connection_args):
    """
    Delete a user (keystone user-delete)

    CLI Examples:

    .. code-block:: bash

        salt '*' keystone.user_delete c965f79c4f864eaaa9c3b41904e67082
        salt '*' keystone.user_delete user_id=c965f79c4f864eaaa9c3b41904e67082
        salt '*' keystone.user_delete name=nova
    """
    kstone = auth(profile, **connection_args)
    if name:
        for user in kstone.users.list():
            if user.name == name:
                user_id = user.id
                break
    if not user_id:
        return {"Error": "Unable to resolve user id"}
    kstone.users.delete(user_id)
    ret = "User ID {} deleted".format(user_id)
    if name:

        ret += " ({})".format(name)
    return ret


def user_update(
    user_id=None,
    name=None,
    email=None,
    enabled=None,
    tenant=None,
    profile=None,
    project=None,
    description=None,
    **connection_args
):
    """
    Update a user's information (keystone user-update)
    The following fields may be updated: name, email, enabled, tenant.
    Because the name is one of the fields, a valid user id is required.

    CLI Examples:

    .. code-block:: bash

        salt '*' keystone.user_update user_id=c965f79c4f864eaaa9c3b41904e67082 name=newname
        salt '*' keystone.user_update c965f79c4f864eaaa9c3b41904e67082 name=newname email=newemail@domain.com
    """
    kstone = auth(profile, **connection_args)
    if not user_id:
        for user in kstone.users.list():
            if user.name == name:
                user_id = user.id
                break
        if not user_id:
            return {"Error": "Unable to resolve user id"}
    user = kstone.users.get(user_id)
    # Keep previous settings if not updating them
    if not name:
        name = user.name
    if not email:
        email = user.email
    if enabled is None:
        enabled = user.enabled

    if _OS_IDENTITY_API_VERSION > 2:
        if description is None:
            description = getattr(user, "description", None)
        else:
            description = str(description)

        project_id = None
        if project:
            for proj in kstone.projects.list():
                if proj.name == project:
                    project_id = proj.id
                    break
        if not project_id:
            project_id = getattr(user, "project_id", None)

        kstone.users.update(
            user=user_id,
            name=name,
            email=email,
            enabled=enabled,
            description=description,
            project_id=project_id,
        )
    else:
        kstone.users.update(user=user_id, name=name, email=email, enabled=enabled)

        tenant_id = None
        if tenant:
            for tnt in kstone.tenants.list():
                if tnt.name == tenant:
                    tenant_id = tnt.id
                    break
            if tenant_id:
                kstone.users.update_tenant(user_id, tenant_id)

    ret = "Info updated for user ID {}".format(user_id)
    return ret


def user_verify_password(
    user_id=None, name=None, password=None, profile=None, **connection_args
):
    """
    Verify a user's password

    CLI Examples:

    .. code-block:: bash

        salt '*' keystone.user_verify_password name=test password=foobar
        salt '*' keystone.user_verify_password user_id=c965f79c4f864eaaa9c3b41904e67082 password=foobar
    """
    kstone = auth(profile, **connection_args)
    if "connection_endpoint" in connection_args:
        auth_url = connection_args.get("connection_endpoint")
    else:
        if _OS_IDENTITY_API_VERSION > 2:
            auth_url = __salt__["config.option"](
                "keystone.endpoint", "http://127.0.0.1:35357/v3"
            )
        else:
            auth_url = __salt__["config.option"](
                "keystone.endpoint", "http://127.0.0.1:35357/v2.0"
            )

    if user_id:
        for user in kstone.users.list():
            if user.id == user_id:
                name = user.name
                break
    if not name:
        return {"Error": "Unable to resolve user name"}
    kwargs = {"username": name, "password": password, "auth_url": auth_url}
    try:
        if _OS_IDENTITY_API_VERSION > 2:
            client3.Client(**kwargs)
        else:
            client.Client(**kwargs)
    except (
        keystoneclient.exceptions.Unauthorized,
        keystoneclient.exceptions.AuthorizationFailure,
    ):
        return False
    return True


def user_password_update(
    user_id=None, name=None, password=None, profile=None, **connection_args
):
    """
    Update a user's password (keystone user-password-update)

    CLI Examples:

    .. code-block:: bash

        salt '*' keystone.user_password_update c965f79c4f864eaaa9c3b41904e67082 password=12345
        salt '*' keystone.user_password_update user_id=c965f79c4f864eaaa9c3b41904e67082 password=12345
        salt '*' keystone.user_password_update name=nova password=12345
    """
    kstone = auth(profile, **connection_args)
    if name:
        for user in kstone.users.list():
            if user.name == name:
                user_id = user.id
                break
    if not user_id:
        return {"Error": "Unable to resolve user id"}

    if _OS_IDENTITY_API_VERSION > 2:
        kstone.users.update(user=user_id, password=password)
    else:
        kstone.users.update_password(user=user_id, password=password)
    ret = "Password updated for user ID {}".format(user_id)
    if name:
        ret += " ({})".format(name)
    return ret


def user_role_add(
    user_id=None,
    user=None,
    tenant_id=None,
    tenant=None,
    role_id=None,
    role=None,
    profile=None,
    project_id=None,
    project_name=None,
    **connection_args
):
    """
    Add role for user in tenant (keystone user-role-add)

    CLI Examples:

    .. code-block:: bash

        salt '*' keystone.user_role_add \
user_id=298ce377245c4ec9b70e1c639c89e654 \
tenant_id=7167a092ece84bae8cead4bf9d15bb3b \
role_id=ce377245c4ec9b70e1c639c89e8cead4
        salt '*' keystone.user_role_add user=admin tenant=admin role=admin
    """
    kstone = auth(profile, **connection_args)

    if project_id and not tenant_id:
        tenant_id = project_id
    elif project_name and not tenant:
        tenant = project_name

    if user:
        user_id = user_get(name=user, profile=profile, **connection_args)[user].get(
            "id"
        )
    else:
        user = next(iter(user_get(user_id, profile=profile, **connection_args).keys()))[
            "name"
        ]
    if not user_id:
        return {"Error": "Unable to resolve user id"}

    if tenant:
        tenant_id = tenant_get(name=tenant, profile=profile, **connection_args)[
            tenant
        ].get("id")
    else:
        tenant = next(
            iter(tenant_get(tenant_id, profile=profile, **connection_args).keys())
        )["name"]
    if not tenant_id:
        return {"Error": "Unable to resolve tenant/project id"}

    if role:
        role_id = role_get(name=role, profile=profile, **connection_args)[role]["id"]
    else:
        role = next(iter(role_get(role_id, profile=profile, **connection_args).keys()))[
            "name"
        ]
    if not role_id:
        return {"Error": "Unable to resolve role id"}

    if _OS_IDENTITY_API_VERSION > 2:
        kstone.roles.grant(role_id, user=user_id, project=tenant_id)
    else:
        kstone.roles.add_user_role(user_id, role_id, tenant_id)
    ret_msg = '"{0}" role added for user "{1}" for "{2}" tenant/project'
    return ret_msg.format(role, user, tenant)


def user_role_remove(
    user_id=None,
    user=None,
    tenant_id=None,
    tenant=None,
    role_id=None,
    role=None,
    profile=None,
    project_id=None,
    project_name=None,
    **connection_args
):
    """
    Remove role for user in tenant (keystone user-role-remove)

    CLI Examples:

    .. code-block:: bash

        salt '*' keystone.user_role_remove \
user_id=298ce377245c4ec9b70e1c639c89e654 \
tenant_id=7167a092ece84bae8cead4bf9d15bb3b \
role_id=ce377245c4ec9b70e1c639c89e8cead4
        salt '*' keystone.user_role_remove user=admin tenant=admin role=admin
    """
    kstone = auth(profile, **connection_args)

    if project_id and not tenant_id:
        tenant_id = project_id
    elif project_name and not tenant:
        tenant = project_name

    if user:
        user_id = user_get(name=user, profile=profile, **connection_args)[user].get(
            "id"
        )
    else:
        user = next(iter(user_get(user_id, profile=profile, **connection_args).keys()))[
            "name"
        ]
    if not user_id:
        return {"Error": "Unable to resolve user id"}

    if tenant:
        tenant_id = tenant_get(name=tenant, profile=profile, **connection_args)[
            tenant
        ].get("id")
    else:
        tenant = next(
            iter(tenant_get(tenant_id, profile=profile, **connection_args).keys())
        )["name"]
    if not tenant_id:
        return {"Error": "Unable to resolve tenant/project id"}

    if role:
        role_id = role_get(name=role, profile=profile, **connection_args)[role]["id"]
    else:
        role = next(iter(role_get(role_id).keys()))["name"]
    if not role_id:
        return {"Error": "Unable to resolve role id"}

    if _OS_IDENTITY_API_VERSION > 2:
        kstone.roles.revoke(role_id, user=user_id, project=tenant_id)
    else:
        kstone.roles.remove_user_role(user_id, role_id, tenant_id)
    ret_msg = '"{0}" role removed for user "{1}" under "{2}" tenant'
    return ret_msg.format(role, user, tenant)


def user_role_list(
    user_id=None,
    tenant_id=None,
    user_name=None,
    tenant_name=None,
    profile=None,
    project_id=None,
    project_name=None,
    **connection_args
):
    """
    Return a list of available user_roles (keystone user-roles-list)

    CLI Examples:

    .. code-block:: bash

        salt '*' keystone.user_role_list \
user_id=298ce377245c4ec9b70e1c639c89e654 \
tenant_id=7167a092ece84bae8cead4bf9d15bb3b
        salt '*' keystone.user_role_list user_name=admin tenant_name=admin
    """
    kstone = auth(profile, **connection_args)
    ret = {}

    if project_id and not tenant_id:
        tenant_id = project_id
    elif project_name and not tenant_name:
        tenant_name = project_name

    if user_name:
        for user in kstone.users.list():
            if user.name == user_name:
                user_id = user.id
                break
    if tenant_name:
        for tenant in getattr(kstone, _TENANTS, None).list():
            if tenant.name == tenant_name:
                tenant_id = tenant.id
                break
    if not user_id or not tenant_id:
        return {"Error": "Unable to resolve user or tenant/project id"}

    if _OS_IDENTITY_API_VERSION > 2:
        for role in kstone.roles.list(user=user_id, project=tenant_id):
            ret[role.name] = {
                value: getattr(role, value)
                for value in dir(role)
                if not value.startswith("_")
                and isinstance(getattr(role, value), (str, dict, bool))
            }
    else:
        for role in kstone.roles.roles_for_user(user=user_id, tenant=tenant_id):
            ret[role.name] = {
                "id": role.id,
                "name": role.name,
                "user_id": user_id,
                "tenant_id": tenant_id,
            }
    return ret


def _item_list(profile=None, **connection_args):
    """
    Template for writing list functions
    Return a list of available items (keystone items-list)

    CLI Example:

    .. code-block:: bash

        salt '*' keystone.item_list
    """
    kstone = auth(profile, **connection_args)
    ret = []
    for item in kstone.items.list():
        ret.append(item.__dict__)
        # ret[item.name] = {
        #         'id': item.id,
        #         'name': item.name,
        #         }
    return ret

    # The following is a list of functions that need to be incorporated in the
    # keystone module. This list should be updated as functions are added.
    #
    # endpoint-create     Create a new endpoint associated with a service
    # endpoint-delete     Delete a service endpoint
    # discover            Discover Keystone servers and show authentication
    #                     protocols and
    # bootstrap           Grants a new role to a new user on a new tenant, after
    #                     creating each.



import os.path

import pytest
import salt.modules.virt as virt
from tests.support.mock import MagicMock, patch

from .conftest import loader_modules_config


@pytest.fixture
def configure_loader_modules():
    return loader_modules_config()


def test_node_devices(make_mock_device):
    """
    Test the virt.node_devices() function
    """
    mock_devs = [
        make_mock_device(
            """
            <device>
              <name>pci_1002_71c4</name>
              <parent>pci_8086_27a1</parent>
              <capability type='pci'>
                <class>0xffffff</class>
                <domain>0</domain>
                <bus>1</bus>
                <slot>0</slot>
                <function>0</function>
                <product id='0x71c4'>M56GL [Mobility FireGL V5200]</product>
                <vendor id='0x1002'>ATI Technologies Inc</vendor>
                <numa node='1'/>
              </capability>
            </device>
        """
        ),
        # Linux USB hub to be ignored
        make_mock_device(
            """
            <device>
              <name>usb_device_1d6b_1_0000_00_1d_0</name>
              <parent>pci_8086_27c8</parent>
              <capability type='usb_device'>
                <bus>2</bus>
                <device>1</device>
                <product id='0x0001'>1.1 root hub</product>
                <vendor id='0x1d6b'>Linux Foundation</vendor>
              </capability>
            </device>
        """
        ),
        # SR-IOV PCI device with multiple capabilities
        make_mock_device(
            """
            <device>
              <name>pci_0000_02_10_7</name>
              <parent>pci_0000_00_04_0</parent>
              <capability type='pci'>
                <domain>0</domain>
                <bus>2</bus>
                <slot>16</slot>
                <function>7</function>
                <product id='0x10ca'>82576 Virtual Function</product>
                <vendor id='0x8086'>Intel Corporation</vendor>
                <capability type='phys_function'>
                  <address domain='0x0000' bus='0x02' slot='0x00' function='0x1'/>
                </capability>
                <capability type='virt_functions' maxCount='7'>
                  <address domain='0x0000' bus='0x02' slot='0x00' function='0x2'/>
                  <address domain='0x0000' bus='0x02' slot='0x00' function='0x3'/>
                  <address domain='0x0000' bus='0x02' slot='0x00' function='0x4'/>
                  <address domain='0x0000' bus='0x02' slot='0x00' function='0x5'/>
                </capability>
                <iommuGroup number='31'>
                  <address domain='0x0000' bus='0x02' slot='0x10' function='0x7'/>
                </iommuGroup>
                <numa node='0'/>
                <pci-express>
                  <link validity='cap' port='0' speed='2.5' width='4'/>
                  <link validity='sta' width='0'/>
                </pci-express>
              </capability>
            </device>
        """
        ),
        # PCI bridge to be ignored
        make_mock_device(
            """
            <device>
              <name>pci_0000_00_1c_0</name>
              <parent>computer</parent>
              <capability type='pci'>
                <class>0xffffff</class>
                <domain>0</domain>
                <bus>0</bus>
                <slot>28</slot>
                <function>0</function>
                <product id='0x8c10'>8 Series/C220 Series Chipset Family PCI Express Root Port #1</product>
                <vendor id='0x8086'>Intel Corporation</vendor>
                <capability type='pci-bridge'/>
                <iommuGroup number='8'>
                  <address domain='0x0000' bus='0x00' slot='0x1c' function='0x0'/>
                </iommuGroup>
                <pci-express>
                  <link validity='cap' port='1' speed='5' width='1'/>
                  <link validity='sta' speed='2.5' width='1'/>
                </pci-express>
              </capability>
            </device>
        """
        ),
        # Other device to be ignored
        make_mock_device(
            """
            <device>
              <name>mdev_3627463d_b7f0_4fea_b468_f1da537d301b</name>
              <parent>computer</parent>
              <capability type='mdev'>
                <type id='mtty-1'/>
                <iommuGroup number='12'/>
              </capability>
            </device>
        """
        ),
        # USB device to be listed
        make_mock_device(
            """
            <device>
              <name>usb_3_1_3</name>
              <path>/sys/devices/pci0000:00/0000:00:1d.6/0000:06:00.0/0000:07:02.0/0000:3e:00.0/usb3/3-1/3-1.3</path>
              <devnode type='dev'>/dev/bus/usb/003/004</devnode>
              <parent>usb_3_1</parent>
              <driver>
                <name>usb</name>
              </driver>
              <capability type='usb_device'>
                <bus>3</bus>
                <device>4</device>
                <product id='0x6006'>AUKEY PC-LM1E Camera</product>
                <vendor id='0x0458'>KYE Systems Corp. (Mouse Systems)</vendor>
              </capability>
            </device>
        """
        ),
        # Network device to be listed
        make_mock_device(
            """
            <device>
              <name>net_eth8_e6_86_48_46_c5_29</name>
              <path>/sys/devices/pci0000:3a/0000:3a:00.0/0000:3b:00.0/0000:3c:03.0/0000:3d:02.2/net/eth8</path>
              <parent>pci_0000_02_10_7</parent>
              <capability type='net'>
                <interface>eth8</interface>
                <address>e6:86:48:46:c5:29</address>
                <link state='down'/>
              </capability>
            </device>
            """
        ),
        # Network device to be ignored
        make_mock_device(
            """
            <device>
              <name>net_lo_00_00_00_00_00_00</name>
              <path>/sys/devices/virtual/net/lo</path>
              <parent>computer</parent>
              <capability type='net'>
                <interface>lo</interface>
                <address>00:00:00:00:00:00</address>
                <link state='unknown'/>
              </capability>
            </device>
            """
        ),
    ]
    virt.libvirt.openAuth().listAllDevices.return_value = mock_devs

    assert virt.node_devices() == [
        {
            "name": "pci_1002_71c4",
            "caps": "pci",
            "vendor_id": "0x1002",
            "vendor": "ATI Technologies Inc",
            "product_id": "0x71c4",
            "product": "M56GL [Mobility FireGL V5200]",
            "address": "0000:01:00.0",
            "PCI class": "0xffffff",
        },
        {
            "name": "pci_0000_02_10_7",
            "caps": "pci",
            "vendor_id": "0x8086",
            "vendor": "Intel Corporation",
            "product_id": "0x10ca",
            "product": "82576 Virtual Function",
            "address": "0000:02:10.7",
            "physical function": "0000:02:00.1",
            "virtual functions": [
                "0000:02:00.2",
                "0000:02:00.3",
                "0000:02:00.4",
                "0000:02:00.5",
            ],
        },
        {
            "name": "usb_3_1_3",
            "caps": "usb_device",
            "vendor": "KYE Systems Corp. (Mouse Systems)",
            "vendor_id": "0x0458",
            "product": "AUKEY PC-LM1E Camera",
            "product_id": "0x6006",
            "address": "003:004",
        },
        {
            "name": "eth8",
            "caps": "net",
            "address": "e6:86:48:46:c5:29",
            "state": "down",
            "device name": "pci_0000_02_10_7",
        },
    ]


@pytest.mark.parametrize(
    "dev_kvm, libvirtd", [(True, True), (False, False), (True, False)]
)
def test_is_kvm(dev_kvm, libvirtd):
    """
    Test the virt._is_kvm_hyper() function
    """
    with patch.dict(os.path.__dict__, {"exists": MagicMock(return_value=dev_kvm)}):
        processes = ["libvirtd"] if libvirtd else []
        with patch.dict(virt.__grains__, {"ps": MagicMock(return_value="foo")}):
            with patch.dict(
                virt.__salt__, {"cmd.run": MagicMock(return_value=processes)}
            ):
                assert virt._is_kvm_hyper() == (dev_kvm and libvirtd)



"""
Test Salt MySQL state module across various MySQL variants
"""
import logging

import pytest
import salt.modules.mysql as mysql
from tests.support.pytest.mysql import mysql_container  # pylint: disable=unused-import

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_if_binaries_missing("dockerd"),
    pytest.mark.skipif(
        mysql.MySQLdb is None, reason="No python mysql client installed."
    ),
]


@pytest.fixture(scope="module")
def salt_cli_wrapper(salt_minion, salt_cli, mysql_container):
    def run_command(*command, **kwargs):
        return salt_cli.run(
            *command,
            minion_tgt=salt_minion.id,
            connection_user=mysql_container.mysql_user,
            connection_pass=mysql_container.mysql_passwd,
            connection_db="mysql",
            connection_port=mysql_container.mysql_port,
            **kwargs
        )

    return run_command


@pytest.fixture(scope="module")
def salt_call_cli_wrapper(salt_call_cli, mysql_container):
    def run_command(*command, **kwargs):
        return salt_call_cli.run(
            *command,
            connection_user=mysql_container.mysql_user,
            connection_pass=mysql_container.mysql_passwd,
            connection_db="mysql",
            connection_port=mysql_container.mysql_port,
            **kwargs
        )

    return run_command


def test_database_present_absent(salt_cli_wrapper):
    ret = salt_cli_wrapper(
        "state.single",
        "mysql_database.present",
        name="test_database",
    )
    state = ret.json["mysql_database_|-test_database_|-test_database_|-present"]
    assert ret.exitcode == 0, ret

    assert "changes" in state
    assert state["changes"] == {"test_database": "Present"}

    assert "comment" in state
    assert state["comment"] == "The database test_database has been created"

    ret = salt_cli_wrapper(
        "state.single",
        "mysql_database.absent",
        name="test_database",
    )
    state = ret.json["mysql_database_|-test_database_|-test_database_|-absent"]

    assert ret.exitcode == 0, ret

    assert "changes" in state
    assert state["changes"] == {"test_database": "Absent"}

    assert "comment" in state
    assert state["comment"] == "Database test_database has been removed"


def test_grants_present_absent(salt_cli_wrapper, salt_call_cli_wrapper):

    # Create the database
    ret = salt_call_cli_wrapper("mysql.db_create", "salt")
    assert ret.json

    # Create a user
    ret = salt_call_cli_wrapper(
        "mysql.user_create",
        "george",
        host="localhost",
        password="badpassword",
    )
    assert ret.json

    ret = salt_cli_wrapper(
        "state.single",
        "mysql_grants.present",
        name="add_salt_grants",
        grant="select,insert,update",
        database="salt.*",
        user="george",
        host="localhost",
    )
    state = ret.json["mysql_grants_|-add_salt_grants_|-add_salt_grants_|-present"]
    assert ret.exitcode == 0, ret

    assert "changes" in state
    assert state["changes"] == {"add_salt_grants": "Present"}

    assert "comment" in state
    assert (
        state["comment"]
        == "Grant select,insert,update on salt.* to george@localhost has been added"
    )

    ret = salt_cli_wrapper(
        "state.single",
        "mysql_grants.absent",
        name="delete_salt_grants",
        grant="select,insert,update",
        database="salt.*",
        user="george",
        host="localhost",
    )
    state = ret.json["mysql_grants_|-delete_salt_grants_|-delete_salt_grants_|-absent"]
    assert ret.exitcode == 0, ret

    assert "changes" in state
    assert state["changes"] == {"delete_salt_grants": "Absent"}

    assert "comment" in state
    assert (
        state["comment"]
        == "Grant select,insert,update on salt.* for george@localhost has been revoked"
    )

    # Remove the user
    ret = salt_call_cli_wrapper("mysql.user_remove", "george", host="localhost")
    assert ret.json

    # Remove the database
    ret = salt_call_cli_wrapper("mysql.db_remove", "salt")
    assert ret.json


def test_user_present_absent(salt_cli_wrapper):

    ret = salt_cli_wrapper(
        "state.single",
        "mysql_user.present",
        name="george",
        host="localhost",
        password="password",
    )
    state = ret.json["mysql_user_|-george_|-george_|-present"]
    assert ret.exitcode == 0, ret

    assert "changes" in state
    assert state["changes"] == {"george": "Present"}

    assert "comment" in state
    assert state["comment"] == "The user george@localhost has been added"

    ret = salt_cli_wrapper(
        "state.single",
        "mysql_user.absent",
        name="george",
        host="localhost",
    )
    state = ret.json["mysql_user_|-george_|-george_|-absent"]
    assert ret.exitcode == 0, ret

    assert "changes" in state
    assert state["changes"] == {"george": "Absent"}

    assert "comment" in state
    assert state["comment"] == "User george@localhost has been removed"


def test_user_present_absent_passwordless(salt_cli_wrapper):

    ret = salt_cli_wrapper(
        "state.single",
        "mysql_user.present",
        name="george",
        host="localhost",
        allow_passwordless=True,
    )
    state = ret.json["mysql_user_|-george_|-george_|-present"]
    assert ret.exitcode == 0, ret

    assert "changes" in state
    assert state["changes"] == {"george": "Present"}

    assert "comment" in state
    log.debug("=== comment %s ===", state["comment"])
    assert (
        state["comment"]
        == "The user george@localhost has been added with passwordless login"
    )

    ret = salt_cli_wrapper(
        "state.single",
        "mysql_user.absent",
        name="george",
        host="localhost",
    )
    state = ret.json["mysql_user_|-george_|-george_|-absent"]
    assert ret.exitcode == 0, ret

    assert "changes" in state
    assert state["changes"] == {"george": "Absent"}

    assert "comment" in state
    assert state["comment"] == "User george@localhost has been removed"


def test_user_present_absent_unixsocket(salt_cli_wrapper, mysql_container):

    # The auth_socket plugin on MariaDB is unavailable
    # on versions 10.1 - 10.3
    if "mariadb" in mysql_container.mysql_name:
        if mysql_container.mysql_version in ("10.1", "10.2", "10.3"):
            pytest.skip(
                "The auth_socket plugin is unavaiable "
                "for the {}:{} docker image.".format(
                    mysql_container.mysql_name, mysql_container.mysql_version
                )
            )

    # enable the auth_socket plugin on MySQL
    # already enabled on MariaDB > 10.3
    if "mariadb" not in mysql_container.mysql_name:
        ret = salt_cli_wrapper("mysql.plugin_add", "auth_socket")
        assert ret.json

    ret = salt_cli_wrapper(
        "state.single",
        "mysql_user.present",
        name="george",
        host="localhost",
        unix_socket=True,
        allow_passwordless=False,
    )
    state = ret.json["mysql_user_|-george_|-george_|-present"]
    assert ret.exitcode == 0, ret

    assert "changes" in state
    assert state["changes"] == {"george": "Present"}

    assert "comment" in state
    assert (
        state["comment"] == "The user george@localhost has been added using unix_socket"
    )

    ret = salt_cli_wrapper(
        "state.single",
        "mysql_user.absent",
        name="george",
        host="localhost",
    )
    state = ret.json["mysql_user_|-george_|-george_|-absent"]
    assert ret.exitcode == 0, ret

    assert "changes" in state
    assert state["changes"] == {"george": "Absent"}

    assert "comment" in state
    assert state["comment"] == "User george@localhost has been removed"

    if "mariadb" not in mysql_container.mysql_name:
        ret = salt_cli_wrapper("mysql.plugin_remove", "auth_socket")
        assert ret.json



"""
tests.integration.modules.pw_user
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

import pytest
from tests.support.case import ModuleCase
from tests.support.helpers import random_string, runs_on


@runs_on(kernel="FreeBSD")
class PwUserModuleTest(ModuleCase):
    @pytest.mark.destructive_test
    @pytest.mark.skip_if_not_root
    def test_groups_includes_primary(self):
        # Let's create a user, which usually creates the group matching the name
        uname = random_string("PWU-", lowercase=False)
        if self.run_function("user.add", [uname]) is not True:
            # Skip because creating is not what we're testing here
            self.run_function("user.delete", [uname, True, True])
            self.skipTest("Failed to create user")

        try:
            uinfo = self.run_function("user.info", [uname])
            self.assertIn(uname, uinfo["groups"])

            # This uid is available, store it
            uid = uinfo["uid"]

            self.run_function("user.delete", [uname, True, True])

            # Now, a weird group id
            gname = random_string("PWU-", lowercase=False)
            if self.run_function("group.add", [gname]) is not True:
                self.run_function("group.delete", [gname, True, True])
                self.skipTest("Failed to create group")

            ginfo = self.run_function("group.info", [gname])

            # And create the user with that gid
            if self.run_function("user.add", [uname, uid, ginfo["gid"]]) is False:
                # Skip because creating is not what we're testing here
                self.run_function("user.delete", [uname, True, True])
                self.skipTest("Failed to create user")

            uinfo = self.run_function("user.info", [uname])
            self.assertIn(gname, uinfo["groups"])

        except AssertionError:
            self.run_function("user.delete", [uname, True, True])
            raise



"""
Manage users with the useradd command

.. important::

    If you feel that Salt should be using this module to manage users on a
    minion, and it is using a different module (or gives an error similar to
    *'user.info' is not available*), see :ref:`here
    <module-provider-override>`.

"""


import copy
import logging

import salt.utils.data
import salt.utils.user
from salt.exceptions import CommandExecutionError

try:
    import pwd

    HAS_PWD = True
except ImportError:
    HAS_PWD = False


log = logging.getLogger(__name__)

# Define the module's virtual name
__virtualname__ = "user"


def __virtual__():
    """
    Set the user module if the kernel is SunOS
    """
    if __grains__["kernel"] == "SunOS" and HAS_PWD:
        return __virtualname__
    return (
        False,
        "The solaris_user execution module failed to load: "
        "only available on Solaris systems with pwd module installed.",
    )


def _get_gecos(name):
    """
    Retrieve GECOS field info and return it in dictionary form
    """
    gecos_field = pwd.getpwnam(name).pw_gecos.split(",", 3)
    if not gecos_field:
        return {}
    else:
        # Assign empty strings for any unspecified trailing GECOS fields
        while len(gecos_field) < 4:
            gecos_field.append("")
        return {
            "fullname": str(gecos_field[0]),
            "roomnumber": str(gecos_field[1]),
            "workphone": str(gecos_field[2]),
            "homephone": str(gecos_field[3]),
        }


def _build_gecos(gecos_dict):
    """
    Accepts a dictionary entry containing GECOS field names and their values,
    and returns a full GECOS comment string, to be used with usermod.
    """
    return "{},{},{},{}".format(
        gecos_dict.get("fullname", ""),
        gecos_dict.get("roomnumber", ""),
        gecos_dict.get("workphone", ""),
        gecos_dict.get("homephone", ""),
    )


def _update_gecos(name, key, value):
    """
    Common code to change a user's GECOS information
    """
    if not isinstance(value, str):
        value = str(value)
    pre_info = _get_gecos(name)
    if not pre_info:
        return False
    if value == pre_info[key]:
        return True
    gecos_data = copy.deepcopy(pre_info)
    gecos_data[key] = value
    cmd = ["usermod", "-c", _build_gecos(gecos_data), name]
    __salt__["cmd.run"](cmd, python_shell=False)
    post_info = info(name)
    return _get_gecos(name).get(key) == value


def add(
    name,
    uid=None,
    gid=None,
    groups=None,
    home=None,
    shell=None,
    unique=True,
    fullname="",
    roomnumber="",
    workphone="",
    homephone="",
    createhome=True,
    **kwargs
):
    """
    Add a user to the minion

    CLI Example:

    .. code-block:: bash

        salt '*' user.add name <uid> <gid> <groups> <home> <shell>
    """
    if salt.utils.data.is_true(kwargs.pop("system", False)):
        log.warning("solaris_user module does not support the 'system' argument")
    if kwargs:
        log.warning("Invalid kwargs passed to user.add")

    if isinstance(groups, str):
        groups = groups.split(",")
    cmd = ["useradd"]
    if shell:
        cmd.extend(["-s", shell])
    if uid:
        cmd.extend(["-u", uid])
    if gid:
        cmd.extend(["-g", gid])
    if groups:
        cmd.extend(["-G", ",".join(groups)])
    if createhome:
        cmd.append("-m")
    if home is not None:
        cmd.extend(["-d", home])
    if not unique:
        cmd.append("-o")
    cmd.append(name)

    if __salt__["cmd.retcode"](cmd, python_shell=False) != 0:
        return False
    else:
        # At this point, the user was successfully created, so return true
        # regardless of the outcome of the below functions. If there is a
        # problem wth changing any of the user's info below, it will be raised
        # in a future highstate call. If anyone has a better idea on how to do
        # this, feel free to change it, but I didn't think it was a good idea
        # to return False when the user was successfully created since A) the
        # user does exist, and B) running useradd again would result in a
        # nonzero exit status and be interpreted as a False result.
        if fullname:
            chfullname(name, fullname)
        if roomnumber:
            chroomnumber(name, roomnumber)
        if workphone:
            chworkphone(name, workphone)
        if homephone:
            chhomephone(name, homephone)
        return True


def delete(name, remove=False, force=False):
    """
    Remove a user from the minion

    CLI Example:

    .. code-block:: bash

        salt '*' user.delete name remove=True force=True
    """
    if salt.utils.data.is_true(force):
        log.warning(
            "userdel does not support force-deleting user while user is logged in"
        )
    cmd = ["userdel"]
    if remove:
        cmd.append("-r")
    cmd.append(name)
    return __salt__["cmd.retcode"](cmd, python_shell=False) == 0


def getent(refresh=False):
    """
    Return the list of all info for all users

    CLI Example:

    .. code-block:: bash

        salt '*' user.getent
    """
    if "user.getent" in __context__ and not refresh:
        return __context__["user.getent"]

    ret = []
    for data in pwd.getpwall():
        ret.append(info(data.pw_name))
    __context__["user.getent"] = ret
    return ret


def chuid(name, uid):
    """
    Change the uid for a named user

    CLI Example:

    .. code-block:: bash

        salt '*' user.chuid foo 4376
    """
    pre_info = info(name)
    if not pre_info:
        raise CommandExecutionError("User '{}' does not exist".format(name))
    if uid == pre_info["uid"]:
        return True
    cmd = ["usermod", "-u", uid, name]
    __salt__["cmd.run"](cmd, python_shell=False)
    return info(name).get("uid") == uid


def chgid(name, gid):
    """
    Change the default group of the user

    CLI Example:

    .. code-block:: bash

        salt '*' user.chgid foo 4376
    """
    pre_info = info(name)
    if not pre_info:
        raise CommandExecutionError("User '{}' does not exist".format(name))
    if gid == pre_info["gid"]:
        return True
    cmd = ["usermod", "-g", gid, name]
    __salt__["cmd.run"](cmd, python_shell=False)
    return info(name).get("gid") == gid


def chshell(name, shell):
    """
    Change the default shell of the user

    CLI Example:

    .. code-block:: bash

        salt '*' user.chshell foo /bin/zsh
    """
    pre_info = info(name)
    if not pre_info:
        raise CommandExecutionError("User '{}' does not exist".format(name))
    if shell == pre_info["shell"]:
        return True
    cmd = ["usermod", "-s", shell, name]
    __salt__["cmd.run"](cmd, python_shell=False)
    return info(name).get("shell") == shell


def chhome(name, home, persist=False):
    """
    Set a new home directory for an existing user

    name
        Username to modify

    home
        New home directory to set

    persist : False
        Set to ``True`` to prevent configuration files in the new home
        directory from being overwritten by the files from the skeleton
        directory.

    CLI Example:

    .. code-block:: bash

        salt '*' user.chhome foo /home/users/foo True
    """
    pre_info = info(name)
    if not pre_info:
        raise CommandExecutionError("User '{}' does not exist".format(name))
    if home == pre_info["home"]:
        return True
    cmd = ["usermod", "-d", home]
    if persist:
        cmd.append("-m")
    cmd.append(name)
    __salt__["cmd.run"](cmd, python_shell=False)
    return info(name).get("home") == home


def chgroups(name, groups, append=False):
    """
    Change the groups to which a user belongs

    name
        Username to modify

    groups
        List of groups to set for the user. Can be passed as a comma-separated
        list or a Python list.

    append : False
        Set to ``True`` to append these groups to the user's existing list of
        groups. Otherwise, the specified groups will replace any existing
        groups for the user.

    CLI Example:

    .. code-block:: bash

        salt '*' user.chgroups foo wheel,root True
    """
    if isinstance(groups, str):
        groups = groups.split(",")
    ugrps = set(list_groups(name))
    if ugrps == set(groups):
        return True
    if append:
        groups.update(ugrps)
    cmd = ["usermod", "-G", ",".join(groups), name]
    return __salt__["cmd.retcode"](cmd, python_shell=False) == 0


def chfullname(name, fullname):
    """
    Change the user's Full Name

    CLI Example:

    .. code-block:: bash

        salt '*' user.chfullname foo "Foo Bar"
    """
    return _update_gecos(name, "fullname", fullname)


def chroomnumber(name, roomnumber):
    """
    Change the user's Room Number

    CLI Example:

    .. code-block:: bash

        salt '*' user.chroomnumber foo 123
    """
    return _update_gecos(name, "roomnumber", roomnumber)


def chworkphone(name, workphone):
    """
    Change the user's Work Phone

    CLI Example:

    .. code-block:: bash

        salt '*' user.chworkphone foo "7735550123"
    """
    return _update_gecos(name, "workphone", workphone)


def chhomephone(name, homephone):
    """
    Change the user's Home Phone

    CLI Example:

    .. code-block:: bash

        salt '*' user.chhomephone foo "7735551234"
    """
    return _update_gecos(name, "homephone", homephone)


def info(name):
    """
    Return user information

    CLI Example:

    .. code-block:: bash

        salt '*' user.info root
    """
    ret = {}
    try:
        data = pwd.getpwnam(name)
        ret["gid"] = data.pw_gid
        ret["groups"] = list_groups(name)
        ret["home"] = data.pw_dir
        ret["name"] = data.pw_name
        ret["passwd"] = data.pw_passwd
        ret["shell"] = data.pw_shell
        ret["uid"] = data.pw_uid
        # Put GECOS info into a list
        gecos_field = data.pw_gecos.split(",", 3)
        # Assign empty strings for any unspecified GECOS fields
        while len(gecos_field) < 4:
            gecos_field.append("")
        ret["fullname"] = gecos_field[0]
        ret["roomnumber"] = gecos_field[1]
        ret["workphone"] = gecos_field[2]
        ret["homephone"] = gecos_field[3]
    except KeyError:
        return {}
    return ret


def list_groups(name):
    """
    Return a list of groups the named user belongs to

    CLI Example:

    .. code-block:: bash

        salt '*' user.list_groups foo
    """
    return salt.utils.user.get_group_list(name)


def list_users():
    """
    Return a list of all users

    CLI Example:

    .. code-block:: bash

        salt '*' user.list_users
    """
    return sorted([user.pw_name for user in pwd.getpwall()])


def rename(name, new_name):
    """
    Change the username for a named user

    CLI Example:

    .. code-block:: bash

        salt '*' user.rename name new_name
    """
    current_info = info(name)
    if not current_info:
        raise CommandExecutionError("User '{}' does not exist".format(name))
    new_info = info(new_name)
    if new_info:
        raise CommandExecutionError("User '{}' already exists".format(new_name))
    cmd = ["usermod", "-l", new_name, name]
    __salt__["cmd.run"](cmd, python_shell=False)
    return info(new_name).get("name") == new_name



import textwrap

import salt.modules.xfs as xfs
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


@patch("salt.modules.xfs._get_mounts", MagicMock(return_value={}))
class XFSTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.xfs
    """

    def setup_loader_modules(self):
        return {xfs: {}}

    def test__blkid_output(self):
        """
        Test xfs._blkid_output when there is data
        """
        blkid_export = textwrap.dedent(
            """
            DEVNAME=/dev/sda1
            UUID=XXXXXXXX-XXXX-XXXX-XXXX-XXXXXXXXXXXX
            TYPE=xfs
            PARTUUID=YYYYYYYY-YY

            DEVNAME=/dev/sdb1
            PARTUUID=ZZZZZZZZ-ZZZZ-ZZZZ-ZZZZ-ZZZZZZZZZZZZ
            """
        )
        # We expect to find only data from /dev/sda1, nothig from
        # /dev/sdb1
        self.assertEqual(
            xfs._blkid_output(blkid_export),
            {
                "/dev/sda1": {
                    "label": None,
                    "partuuid": "YYYYYYYY-YY",
                    "uuid": "XXXXXXXX-XXXX-XXXX-XXXX-XXXXXXXXXXXX",
                }
            },
        )



"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>
"""

import logging

import pytest  # pylint: disable=unused-import
import salt.exceptions
import salt.state
import salt.utils.files
import salt.utils.platform
from tests.support.mock import MagicMock, patch
from tests.support.unit import skipIf

log = logging.getLogger(__name__)


@pytest.fixture
def state_obj():
    with patch("salt.state.State._gather_pillar"):
        minion_opts = salt.config.DEFAULT_MINION_OPTS.copy()
        yield salt.state.State(minion_opts)


def test_format_slots_no_slots(state_obj):
    """
    Test the format slots keeps data without slots untouched.
    """
    cdata = {"args": ["arg"], "kwargs": {"key": "val"}}
    state_obj.format_slots(cdata)
    assert cdata == {"args": ["arg"], "kwargs": {"key": "val"}}


@pytest.mark.slow_test
def test_format_slots_arg(state_obj):
    """
    Test the format slots is calling a slot specified in args with corresponding arguments.
    """
    cdata = {
        "args": ["__slot__:salt:mod.fun(fun_arg, fun_key=fun_val)"],
        "kwargs": {"key": "val"},
    }
    mock = MagicMock(return_value="fun_return")
    with patch.dict(state_obj.functions, {"mod.fun": mock}):
        state_obj.format_slots(cdata)
    mock.assert_called_once_with("fun_arg", fun_key="fun_val")
    assert cdata == {"args": ["fun_return"], "kwargs": {"key": "val"}}


@pytest.mark.slow_test
def test_format_slots_dict_arg(state_obj):
    """
    Test the format slots is calling a slot specified in dict arg.
    """
    cdata = {
        "args": [{"subarg": "__slot__:salt:mod.fun(fun_arg, fun_key=fun_val)"}],
        "kwargs": {"key": "val"},
    }
    mock = MagicMock(return_value="fun_return")
    with patch.dict(state_obj.functions, {"mod.fun": mock}):
        state_obj.format_slots(cdata)
    mock.assert_called_once_with("fun_arg", fun_key="fun_val")
    assert cdata == {"args": [{"subarg": "fun_return"}], "kwargs": {"key": "val"}}


@pytest.mark.slow_test
def test_format_slots_listdict_arg(state_obj):
    """
    Test the format slots is calling a slot specified in list containing a dict.
    """
    cdata = {
        "args": [[{"subarg": "__slot__:salt:mod.fun(fun_arg, fun_key=fun_val)"}]],
        "kwargs": {"key": "val"},
    }
    mock = MagicMock(return_value="fun_return")
    with patch.dict(state_obj.functions, {"mod.fun": mock}):
        state_obj.format_slots(cdata)
    mock.assert_called_once_with("fun_arg", fun_key="fun_val")
    assert cdata == {"args": [[{"subarg": "fun_return"}]], "kwargs": {"key": "val"}}


@pytest.mark.slow_test
def test_format_slots_liststr_arg(state_obj):
    """
    Test the format slots is calling a slot specified in list containing a dict.
    """
    cdata = {
        "args": [["__slot__:salt:mod.fun(fun_arg, fun_key=fun_val)"]],
        "kwargs": {"key": "val"},
    }
    mock = MagicMock(return_value="fun_return")
    with patch.dict(state_obj.functions, {"mod.fun": mock}):
        state_obj.format_slots(cdata)
    mock.assert_called_once_with("fun_arg", fun_key="fun_val")
    assert cdata == {"args": [["fun_return"]], "kwargs": {"key": "val"}}


@pytest.mark.slow_test
def test_format_slots_kwarg(state_obj):
    """
    Test the format slots is calling a slot specified in kwargs with corresponding arguments.
    """
    cdata = {
        "args": ["arg"],
        "kwargs": {"key": "__slot__:salt:mod.fun(fun_arg, fun_key=fun_val)"},
    }
    mock = MagicMock(return_value="fun_return")
    with patch.dict(state_obj.functions, {"mod.fun": mock}):
        state_obj.format_slots(cdata)
    mock.assert_called_once_with("fun_arg", fun_key="fun_val")
    assert cdata == {"args": ["arg"], "kwargs": {"key": "fun_return"}}


@pytest.mark.slow_test
def test_format_slots_multi(state_obj):
    """
    Test the format slots is calling all slots with corresponding arguments when multiple slots
    specified.
    """
    cdata = {
        "args": [
            "__slot__:salt:test_mod.fun_a(a_arg, a_key=a_kwarg)",
            "__slot__:salt:test_mod.fun_b(b_arg, b_key=b_kwarg)",
        ],
        "kwargs": {
            "kw_key_1": "__slot__:salt:test_mod.fun_c(c_arg, c_key=c_kwarg)",
            "kw_key_2": "__slot__:salt:test_mod.fun_d(d_arg, d_key=d_kwarg)",
        },
    }
    mock_a = MagicMock(return_value="fun_a_return")
    mock_b = MagicMock(return_value="fun_b_return")
    mock_c = MagicMock(return_value="fun_c_return")
    mock_d = MagicMock(return_value="fun_d_return")
    with patch.dict(
        state_obj.functions,
        {
            "test_mod.fun_a": mock_a,
            "test_mod.fun_b": mock_b,
            "test_mod.fun_c": mock_c,
            "test_mod.fun_d": mock_d,
        },
    ):
        state_obj.format_slots(cdata)
    mock_a.assert_called_once_with("a_arg", a_key="a_kwarg")
    mock_b.assert_called_once_with("b_arg", b_key="b_kwarg")
    mock_c.assert_called_once_with("c_arg", c_key="c_kwarg")
    mock_d.assert_called_once_with("d_arg", d_key="d_kwarg")
    assert cdata == {
        "args": ["fun_a_return", "fun_b_return"],
        "kwargs": {"kw_key_1": "fun_c_return", "kw_key_2": "fun_d_return"},
    }


@pytest.mark.slow_test
def test_format_slots_malformed(state_obj):
    """
    Test the format slots keeps malformed slots untouched.
    """
    sls_data = {
        "args": [
            "__slot__:NOT_SUPPORTED:not.called()",
            "__slot__:salt:not.called(",
            "__slot__:salt:",
            "__slot__:salt",
            "__slot__:",
            "__slot__",
        ],
        "kwargs": {
            "key3": "__slot__:NOT_SUPPORTED:not.called()",
            "key4": "__slot__:salt:not.called(",
            "key5": "__slot__:salt:",
            "key6": "__slot__:salt",
            "key7": "__slot__:",
            "key8": "__slot__",
        },
    }
    cdata = sls_data.copy()
    mock = MagicMock(return_value="return")
    with patch.dict(state_obj.functions, {"not.called": mock}):
        state_obj.format_slots(cdata)
    mock.assert_not_called()
    assert cdata == sls_data


@pytest.mark.slow_test
def test_slot_traverse_dict(state_obj):
    """
    Test the slot parsing of dict response.
    """
    cdata = {
        "args": ["arg"],
        "kwargs": {"key": "__slot__:salt:mod.fun(fun_arg, fun_key=fun_val).key1"},
    }
    return_data = {"key1": "value1"}
    mock = MagicMock(return_value=return_data)
    with patch.dict(state_obj.functions, {"mod.fun": mock}):
        state_obj.format_slots(cdata)
    mock.assert_called_once_with("fun_arg", fun_key="fun_val")
    assert cdata == {"args": ["arg"], "kwargs": {"key": "value1"}}


@pytest.mark.slow_test
def test_slot_append(state_obj):
    """
    Test the slot parsing of dict response.
    """
    cdata = {
        "args": ["arg"],
        "kwargs": {
            "key": "__slot__:salt:mod.fun(fun_arg, fun_key=fun_val).key1 ~ thing~",
        },
    }
    return_data = {"key1": "value1"}
    mock = MagicMock(return_value=return_data)
    with patch.dict(state_obj.functions, {"mod.fun": mock}):
        state_obj.format_slots(cdata)
    mock.assert_called_once_with("fun_arg", fun_key="fun_val")
    assert cdata == {"args": ["arg"], "kwargs": {"key": "value1thing~"}}


# Skip on windows like integration.modules.test_state.StateModuleTest.test_parallel_state_with_long_tag
@skipIf(
    salt.utils.platform.is_windows(),
    "Skipped until parallel states can be fixed on Windows",
)
def test_format_slots_parallel(state_obj):
    """
    Test if slots work with "parallel: true".
    """
    high_data = {
        "always-changes-and-succeeds": {
            "test": [
                {"changes": True},
                {"comment": "__slot__:salt:test.echo(fun_return)"},
                {"parallel": True},
                "configurable_test_state",
                {"order": 10000},
            ],
            "__env__": "base",
            "__sls__": "parallel_slots",
        }
    }
    state_obj.jid = "123"
    res = state_obj.call_high(high_data)
    state_obj.jid = None
    [(_, data)] = res.items()
    assert data["comment"] == "fun_return"



"""
Network Probes
===============

Configure RPM (JunOS)/SLA (Cisco) probes on the device via NAPALM proxy.

:codeauthor: Mircea Ulinic <ping@mirceaulinic.net> & Jerome Fleury <jf@cloudflare.com>
:maturity:   new
:depends:    napalm
:platform:   unix

Dependencies
------------

- :mod:`napalm probes management module <salt.modules.napalm_probes>`

.. versionadded:: 2016.11.0
"""

import copy
import logging

import salt.utils.json
import salt.utils.napalm

log = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------------------------------------
# state properties
# ----------------------------------------------------------------------------------------------------------------------

__virtualname__ = "probes"

# ----------------------------------------------------------------------------------------------------------------------
# global variables
# ----------------------------------------------------------------------------------------------------------------------

# ----------------------------------------------------------------------------------------------------------------------
# property functions
# ----------------------------------------------------------------------------------------------------------------------


def __virtual__():
    """
    NAPALM library must be installed for this module to work and run in a (proxy) minion.
    """
    return salt.utils.napalm.virtual(__opts__, __virtualname__, __file__)


# ----------------------------------------------------------------------------------------------------------------------
# helper functions -- will not be exported
# ----------------------------------------------------------------------------------------------------------------------


def _default_ret(name):

    """
    Returns a default structure of the dictionary to be returned as output of the state functions.
    """

    return {"name": name, "result": False, "changes": {}, "comment": ""}


def _retrieve_rpm_probes():

    """
    Will retrieve the probes from the network device using salt module "probes" throught NAPALM proxy.
    """

    return __salt__["probes.config"]()


def _expand_probes(probes, defaults):

    """
    Updates the probes dictionary with different levels of default values.
    """

    expected_probes = {}

    for probe_name, probe_test in probes.items():
        if probe_name not in expected_probes.keys():
            expected_probes[probe_name] = {}
        probe_defaults = probe_test.pop("defaults", {})
        for test_name, test_details in probe_test.items():
            test_defaults = test_details.pop("defaults", {})
            expected_test_details = copy.deepcopy(
                defaults
            )  # copy first the general defaults
            expected_test_details.update(
                probe_defaults
            )  # update with more specific defaults if any
            expected_test_details.update(
                test_defaults
            )  # update with the most specific defaults if possible
            expected_test_details.update(
                test_details
            )  # update with the actual config of the test
            if test_name not in expected_probes[probe_name].keys():
                expected_probes[probe_name][test_name] = expected_test_details

    return expected_probes


def _clean_probes(probes):

    """
    Will remove empty and useless values from the probes dictionary.
    """

    probes = _ordered_dict_to_dict(
        probes
    )  # make sure we are working only with dict-type
    probes_copy = copy.deepcopy(probes)
    for probe_name, probe_tests in probes_copy.items():
        if not probe_tests:
            probes.pop(probe_name)
            continue
        for test_name, test_params in probe_tests.items():
            if not test_params:
                probes[probe_name].pop(test_name)
            if not probes.get(probe_name):
                probes.pop(probe_name)

    return True


def _compare_probes(configured_probes, expected_probes):

    """
    Compares configured probes on the device with the expected configuration and returns the differences.
    """

    new_probes = {}
    update_probes = {}
    remove_probes = {}

    # noth configured => configure with expected probes
    if not configured_probes:
        return {"add": expected_probes}

    # noting expected => remove everything
    if not expected_probes:
        return {"remove": configured_probes}

    configured_probes_keys_set = set(configured_probes.keys())
    expected_probes_keys_set = set(expected_probes.keys())
    new_probes_keys_set = expected_probes_keys_set - configured_probes_keys_set
    remove_probes_keys_set = configured_probes_keys_set - expected_probes_keys_set

    # new probes
    for probe_name in new_probes_keys_set:
        new_probes[probe_name] = expected_probes.pop(probe_name)

    # old probes, to be removed
    for probe_name in remove_probes_keys_set:
        remove_probes[probe_name] = configured_probes.pop(probe_name)

    # common probes
    for probe_name, probe_tests in expected_probes.items():
        configured_probe_tests = configured_probes.get(probe_name, {})
        configured_tests_keys_set = set(configured_probe_tests.keys())
        expected_tests_keys_set = set(probe_tests.keys())
        new_tests_keys_set = expected_tests_keys_set - configured_tests_keys_set
        remove_tests_keys_set = configured_tests_keys_set - expected_tests_keys_set

        # new tests for common probes
        for test_name in new_tests_keys_set:
            if probe_name not in new_probes.keys():
                new_probes[probe_name] = {}
            new_probes[probe_name].update({test_name: probe_tests.pop(test_name)})
        # old tests for common probes
        for test_name in remove_tests_keys_set:
            if probe_name not in remove_probes.keys():
                remove_probes[probe_name] = {}
            remove_probes[probe_name].update(
                {test_name: configured_probe_tests.pop(test_name)}
            )
        # common tests for common probes
        for test_name, test_params in probe_tests.items():
            configured_test_params = configured_probe_tests.get(test_name, {})
            # if test params are different, probe goes to update probes dict!
            if test_params != configured_test_params:
                if probe_name not in update_probes.keys():
                    update_probes[probe_name] = {}
                update_probes[probe_name].update({test_name: test_params})

    return {"add": new_probes, "update": update_probes, "remove": remove_probes}


def _ordered_dict_to_dict(probes):

    """Mandatory to be dict type in order to be used in the NAPALM Jinja template."""

    return salt.utils.json.loads(salt.utils.json.dumps(probes))


def _set_rpm_probes(probes):

    """
    Calls the Salt module "probes" to configure the probes on the device.
    """

    return __salt__["probes.set_probes"](
        _ordered_dict_to_dict(probes),  # make sure this does not contain ordered dicts
        commit=False,
    )


def _schedule_probes(probes):

    """
    Calls the Salt module "probes" to schedule the configured probes on the device.
    """

    return __salt__["probes.schedule_probes"](
        _ordered_dict_to_dict(probes),  # make sure this does not contain ordered dicts
        commit=False,
    )


def _delete_rpm_probes(probes):

    """
    Calls the Salt module "probes" to delete probes from the device.
    """

    return __salt__["probes.delete_probes"](
        _ordered_dict_to_dict(
            probes
        ),  # not mandatory, but let's make sure we catch all cases
        commit=False,
    )


# ----------------------------------------------------------------------------------------------------------------------
# callable functions
# ----------------------------------------------------------------------------------------------------------------------


def managed(name, probes, defaults=None):

    """
    Ensure the networks device is configured as specified in the state SLS file.
    Probes not specified will be removed, while probes not confiured as expected will trigger config updates.

    :param probes: Defines the probes as expected to be configured on the
        device.  In order to ease the configuration and avoid repeating the
        same parameters for each probe, the next parameter (defaults) can be
        used, providing common characteristics.

    :param defaults: Specifies common parameters for the probes.

    SLS Example:

    .. code-block:: yaml

        rpmprobes:
            probes.managed:
                - probes:
                    probe_name1:
                        probe1_test1:
                            source: 192.168.0.2
                            target: 192.168.0.1
                        probe1_test2:
                            target: 172.17.17.1
                        probe1_test3:
                            target: 8.8.8.8
                            probe_type: http-ping
                    probe_name2:
                        probe2_test1:
                            test_interval: 100
                - defaults:
                    target: 10.10.10.10
                    probe_count: 15
                    test_interval: 3
                    probe_type: icmp-ping

    In the probes configuration, the only mandatory attribute is *target*
    (specified either in probes configuration, either in the defaults
    dictionary).  All the other parameters will use the operating system
    defaults, if not provided:

    - ``source`` - Specifies the source IP Address to be used during the tests.  If
      not specified will use the IP Address of the logical interface loopback0.

    - ``target`` - Destination IP Address.
    - ``probe_count`` - Total number of probes per test (1..15). System
      defaults: 1 on both JunOS & Cisco.
    - ``probe_interval`` - Delay between tests (0..86400 seconds). System
      defaults: 3 on JunOS, 5 on Cisco.
    - ``probe_type`` - Probe request type. Available options:

      - icmp-ping
      - tcp-ping
      - udp-ping

    Using the example configuration above, after running the state, on the device will be configured 4 probes,
    with the following properties:

    .. code-block:: yaml

        probe_name1:
            probe1_test1:
                source: 192.168.0.2
                target: 192.168.0.1
                probe_count: 15
                test_interval: 3
                probe_type: icmp-ping
            probe1_test2:
                target: 172.17.17.1
                probe_count: 15
                test_interval: 3
                probe_type: icmp-ping
            probe1_test3:
                target: 8.8.8.8
                probe_count: 15
                test_interval: 3
                probe_type: http-ping
        probe_name2:
            probe2_test1:
                target: 10.10.10.10
                probe_count: 15
                test_interval: 3
                probe_type: icmp-ping
    """

    ret = _default_ret(name)

    result = True
    comment = ""

    rpm_probes_config = (
        _retrieve_rpm_probes()
    )  # retrieves the RPM config from the device
    if not rpm_probes_config.get("result"):
        ret.update(
            {
                "result": False,
                "comment": (
                    "Cannot retrieve configurtion of the probes from the device:"
                    " {reason}".format(reason=rpm_probes_config.get("comment"))
                ),
            }
        )
        return ret

    # build expect probes config dictionary
    # using default values
    configured_probes = rpm_probes_config.get("out", {})
    if not isinstance(defaults, dict):
        defaults = {}
    expected_probes = _expand_probes(probes, defaults)

    _clean_probes(
        configured_probes
    )  # let's remove the unnecessary data from the configured probes
    _clean_probes(expected_probes)  # also from the expected data

    # ----- Compare expected config with the existing config ---------------------------------------------------------->

    diff = _compare_probes(configured_probes, expected_probes)  # compute the diff

    # <---- Compare expected config with the existing config -----------------------------------------------------------

    # ----- Call set_probes and delete_probes as needed --------------------------------------------------------------->

    add_probes = diff.get("add")
    update_probes = diff.get("update")
    remove_probes = diff.get("remove")

    changes = {
        "added": _ordered_dict_to_dict(add_probes),
        "updated": _ordered_dict_to_dict(update_probes),
        "removed": _ordered_dict_to_dict(remove_probes),
    }

    ret.update({"changes": changes})

    if __opts__["test"] is True:
        ret.update(
            {"comment": "Testing mode: configuration was not changed!", "result": None}
        )
        return ret

    config_change_expected = (
        False  # to check if something changed and a commit would be needed
    )

    if add_probes:
        added = _set_rpm_probes(add_probes)
        if added.get("result"):
            config_change_expected = True
        else:
            result = False
            comment += "Cannot define new probes: {reason}\n".format(
                reason=added.get("comment")
            )

    if update_probes:
        updated = _set_rpm_probes(update_probes)
        if updated.get("result"):
            config_change_expected = True
        else:
            result = False
            comment += "Cannot update probes: {reason}\n".format(
                reason=updated.get("comment")
            )

    if remove_probes:
        removed = _delete_rpm_probes(remove_probes)
        if removed.get("result"):
            config_change_expected = True
        else:
            result = False
            comment += "Cannot remove probes! {reason}\n".format(
                reason=removed.get("comment")
            )

    # <---- Call set_probes and delete_probes as needed ----------------------------------------------------------------

    # ----- Try to save changes --------------------------------------------------------------------------------------->

    if config_change_expected:
        # if any changes expected, try to commit
        result, comment = __salt__["net.config_control"]()

    # <---- Try to save changes ----------------------------------------------------------------------------------------

    # ----- Try to schedule the probes -------------------------------------------------------------------------------->

    add_scheduled = _schedule_probes(add_probes)
    if add_scheduled.get("result"):
        # if able to load the template to schedule the probes, try to commit the scheduling data
        # (yes, a second commit is needed)
        # on devices such as Juniper, RPM probes do not need to be scheduled
        # therefore the template is empty and won't try to commit empty changes
        result, comment = __salt__["net.config_control"]()

    if config_change_expected:
        if result and comment == "":  # if any changes and was able to apply them
            comment = "Probes updated successfully!"

    ret.update({"result": result, "comment": comment})

    return ret
